# pylint: disable-all
try:
    import qpd_pandas  # noqa: F401

    HAS_QPD = True
except ImportError:  # pragma: no cover
    HAS_QPD = False

import copy
import os
import pickle
from datetime import datetime
from unittest import TestCase

import pandas as pd
import pytest
from pytest import raises
from triad.collections.fs import FileSystem
from triad.exceptions import InvalidOperationError

import fugue.api as fa
import fugue.column.functions as ff
from fugue import (
    ArrayDataFrame,
    DataFrame,
    DataFrames,
    ExecutionEngine,
    PandasDataFrame,
    PartitionSpec,
    register_default_sql_engine,
)
from fugue.column import all_cols, col, lit
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.execution.native_execution_engine import NativeExecutionEngine


class ExecutionEngineTests(object):
    """ExecutionEngine level general test suite.
    Any new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    should pass this test suite.
    """

    class Tests(TestCase):
        @classmethod
        def setUpClass(cls):
            register_default_sql_engine(lambda engine: engine.sql_engine)
            cls._engine = cls.make_engine(cls)
            fa.set_global_engine(cls._engine)

        @property
        def engine(self) -> ExecutionEngine:
            return self._engine  # type: ignore

        @classmethod
        def tearDownClass(cls):
            fa.clear_global_engine()
            cls._engine.stop()

        def make_engine(self) -> ExecutionEngine:  # pragma: no cover
            raise NotImplementedError

        def test_init(self):
            print(self.engine)
            assert self.engine.log is not None
            assert self.engine.fs is not None
            assert copy.copy(self.engine) is self.engine
            assert copy.deepcopy(self.engine) is self.engine

        def test_get_parallelism(self):
            assert fa.get_current_parallelism() == 1

        def test_to_df_general(self):
            e = self.engine
            o = ArrayDataFrame(
                [[1.1, 2.2], [3.3, 4.4]],
                "a:double,b:double",
            )
            # all engines should accept these types of inputs
            # should take fugue.DataFrame
            df_eq(o, fa.as_fugue_engine_df(e, o), throw=True)
            # should take array, shema
            df_eq(
                o,
                fa.as_fugue_engine_df(e, [[1.1, 2.2], [3.3, 4.4]], "a:double,b:double"),
                throw=True,
            )
            # should take pandas dataframe
            pdf = pd.DataFrame([[1.1, 2.2], [3.3, 4.4]], columns=["a", "b"])
            df_eq(o, fa.as_fugue_engine_df(e, pdf), throw=True)

            # should convert string to datetime in to_df
            df_eq(
                fa.as_fugue_engine_df(e, [["2020-01-01"]], "a:datetime"),
                [[datetime(2020, 1, 1)]],
                "a:datetime",
                throw=True,
            )

            # should handle empty pandas dataframe
            o = ArrayDataFrame([], "a:double,b:str")
            pdf = pd.DataFrame([[0.1, "a"]], columns=["a", "b"])
            pdf = pdf[pdf.a < 0]
            df_eq(o, fa.as_fugue_engine_df(e, pdf), throw=True)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_filter(self):
            a = ArrayDataFrame(
                [[1, 2], [None, 2], [None, 1], [3, 4], [None, 4]],
                "a:double,b:int",
            )
            b = fa.filter(a, col("a").not_null())
            df_eq(b, [[1, 2], [3, 4]], "a:double,b:int", throw=True)
            c = fa.filter(a, col("a").not_null() & (col("b") < 3))
            df_eq(c, [[1, 2]], "a:double,b:int", throw=True)
            c = fa.filter(a, col("a") + col("b") == 3)
            df_eq(c, [[1, 2]], "a:double,b:int", throw=True)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_select(self):
            a = ArrayDataFrame(
                [[1, 2], [None, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int"
            )

            # simple
            b = fa.select(a, col("b"), (col("b") + 1).alias("c").cast(str))
            df_eq(
                b,
                [[2, "3"], [2, "3"], [1, "2"], [4, "5"], [4, "5"]],
                "b:int,c:str",
                throw=True,
            )

            # with distinct
            b = fa.select(
                a, col("b"), (col("b") + 1).alias("c").cast(str), distinct=True
            )
            df_eq(
                b,
                [[2, "3"], [1, "2"], [4, "5"]],
                "b:int,c:str",
                throw=True,
            )

            # wildcard
            b = fa.select(a, all_cols(), where=col("a") + col("b") == 3)
            df_eq(b, [[1, 2]], "a:double,b:int", throw=True)

            # aggregation
            b = fa.select(a, col("a"), ff.sum(col("b")).cast(float).alias("b"))
            df_eq(b, [[1, 2], [3, 4], [None, 7]], "a:double,b:double", throw=True)

            # having
            # https://github.com/fugue-project/fugue/issues/222
            col_b = ff.sum(col("b"))
            b = fa.select(
                a,
                col("a"),
                col_b.cast(float).alias("c"),
                having=(col_b >= 7) | (col("a") == 1),
            )
            df_eq(b, [[1, 2], [None, 7]], "a:double,c:double", throw=True)

            # literal + alias inference
            # https://github.com/fugue-project/fugue/issues/222
            col_b = ff.sum(col("b"))
            b = fa.select(
                a,
                col("a"),
                lit(1, "o").cast(str),
                col_b.cast(float).alias("c"),
                having=(col_b >= 7) | (col("a") == 1),
            )
            df_eq(
                b, [[1, "1", 2], [None, "1", 7]], "a:double,o:str,c:double", throw=True
            )

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_assign(self):
            a = ArrayDataFrame(
                [[1, 2], [None, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int"
            )

            b = fa.assign(a, x=1, b=col("b").cast(str), c=(col("b") + 1).cast(int))
            df_eq(
                b,
                [
                    [1, "2", 1, 3],
                    [None, "2", 1, 3],
                    [None, "1", 1, 2],
                    [3, "4", 1, 5],
                    [None, "4", 1, 5],
                ],
                "a:double,b:str,x:long,c:long",
                throw=True,
            )

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_aggregate(self):
            a = ArrayDataFrame(
                [[1, 2], [None, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int"
            )

            b = fa.aggregate(
                df=a,
                b=ff.max(col("b")),
                c=(ff.max(col("b")) * 2).cast("int32").alias("c"),
            )
            df_eq(b, [[4, 8]], "b:int,c:int", throw=True)

            b = fa.aggregate(
                a,
                "a",
                b=ff.max(col("b")),
                c=(ff.max(col("b")) * 2).cast("int32").alias("c"),
            )
            df_eq(
                b,
                [[None, 4, 8], [1, 2, 4], [3, 4, 8]],
                "a:double,b:int,c:int",
                throw=True,
            )

            with raises(ValueError):
                fa.aggregate(a, "a", b=ff.max(col("b")), x=1)

            with raises(ValueError):
                fa.aggregate(a, "a")

        def test_map(self):
            def noop(cursor, data):
                return data

            def on_init(partition_no, data):
                # TODO: this test is not sufficient
                assert partition_no >= 0
                data.peek_array()

            e = self.engine
            o = ArrayDataFrame(
                [[1, 2], [None, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int"
            )
            a = fa.as_fugue_engine_df(e, o)
            # no partition
            c = e.map_engine.map_dataframe(a, noop, a.schema, PartitionSpec())
            df_eq(c, o, throw=True)
            # with key partition
            c = e.map_engine.map_dataframe(
                a, noop, a.schema, PartitionSpec(by=["a"], presort="b")
            )
            df_eq(c, o, throw=True)
            # select top
            c = e.map_engine.map_dataframe(
                a, select_top, a.schema, PartitionSpec(by=["a"], presort="b")
            )
            df_eq(c, [[None, 1], [1, 2], [3, 4]], "a:double,b:int", throw=True)
            # select top with another order
            c = e.map_engine.map_dataframe(
                a,
                select_top,
                a.schema,
                PartitionSpec(partition_by=["a"], presort="b DESC"),
            )
            df_eq(
                c,
                [[None, 4], [1, 2], [3, 4]],
                "a:double,b:int",
                throw=True,
            )
            # add num_partitions, on_init should not matter
            c = e.map_engine.map_dataframe(
                a,
                select_top,
                a.schema,
                PartitionSpec(partition_by=["a"], presort="b DESC", num_partitions=3),
                on_init=on_init,
            )
            df_eq(c, [[None, 4], [1, 2], [3, 4]], "a:double,b:int", throw=True)

        def test_map_with_special_values(self):
            def with_nat(cursor, data):
                df = data.as_pandas()
                df["nat"] = pd.NaT
                schema = data.schema + "nat:datetime"
                return PandasDataFrame(df, schema)

            e = self.engine
            # test with multiple key with null values
            o = ArrayDataFrame(
                [[1, None, 1], [1, None, 0], [None, None, 1]], "a:double,b:double,c:int"
            )
            c = e.map_engine.map_dataframe(
                o, select_top, o.schema, PartitionSpec(by=["a", "b"], presort="c")
            )
            df_eq(
                c,
                [[1, None, 0], [None, None, 1]],
                "a:double,b:double,c:int",
                throw=True,
            )
            # test datetime with nat
            dt = datetime.now()
            o = ArrayDataFrame(
                [
                    [dt, 2, 1],
                    [None, 2, None],
                    [None, 1, None],
                    [dt, 5, 1],
                    [None, 4, None],
                ],
                "a:datetime,b:int,c:double",
            )
            c = e.map_engine.map_dataframe(
                o, select_top, o.schema, PartitionSpec(by=["a", "c"], presort="b DESC")
            )
            df_eq(
                c,
                [[None, 4, None], [dt, 5, 1]],
                "a:datetime,b:int,c:double",
                throw=True,
            )
            d = e.map_engine.map_dataframe(
                c, with_nat, "a:datetime,b:int,c:double,nat:datetime", PartitionSpec()
            )
            df_eq(
                d,
                [[None, 4, None, None], [dt, 5, 1, None]],
                "a:datetime,b:int,c:double,nat:datetime",
                throw=True,
            )
            # test list
            o = ArrayDataFrame([[dt, [1, 2]]], "a:datetime,b:[int]")
            c = e.map_engine.map_dataframe(
                o, select_top, o.schema, PartitionSpec(by=["a"])
            )
            df_eq(c, o, check_order=True, throw=True)

        def test_map_with_dict_col(self):
            e = self.engine
            dt = datetime.now()
            # test dict
            o = ArrayDataFrame([[dt, dict(a=1)]], "a:datetime,b:{a:int}")
            c = e.map_engine.map_dataframe(
                o, select_top, o.schema, PartitionSpec(by=["a"])
            )
            df_eq(c, o, no_pandas=True, check_order=True, throw=True)

        def test_map_with_binary(self):
            e = self.engine
            o = ArrayDataFrame(
                [[pickle.dumps(BinaryObject("a"))], [pickle.dumps(BinaryObject("b"))]],
                "a:bytes",
            )
            c = e.map_engine.map_dataframe(o, binary_map, o.schema, PartitionSpec())
            expected = ArrayDataFrame(
                [
                    [pickle.dumps(BinaryObject("ax"))],
                    [pickle.dumps(BinaryObject("bx"))],
                ],
                "a:bytes",
            )
            df_eq(expected, c, no_pandas=True, check_order=True, throw=True)

        def test_join_multiple(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[1, 20], [3, 40]], "a:int,c:int")
            c = fa.as_fugue_engine_df(e, [[1, 200], [3, 400]], "a:int,d:int")
            d = fa.inner_join(a, b, c)
            df_eq(
                d,
                [[1, 2, 20, 200], [3, 4, 40, 400]],
                "a:int,b:int,c:int,d:int",
                throw=True,
            )

        def test__join_cross(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6], [7]], "c:int")
            c = fa.join(a, b, how="Cross")
            df_eq(
                c,
                [[1, 2, 6], [1, 2, 7], [3, 4, 6], [3, 4, 7]],
                "a:int,b:int,c:int",
                throw=True,
            )

            b = fa.as_fugue_engine_df(e, [], "c:int")
            c = fa.cross_join(a, b)
            df_eq(c, [], "a:int,b:int,c:int", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [], "c:int")
            c = fa.join(a, b, how="Cross")
            df_eq(c, [], "a:int,b:int,c:int", throw=True)

        def test__join_inner(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.join(a, b, how="INNER", on=["a"])
            df_eq(c, [[1, 2, 6]], "a:int,b:int,c:int", throw=True)
            c = fa.inner_join(b, a)
            df_eq(c, [[6, 1, 2]], "c:int,a:int,b:int", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.join(a, b, how="INNER", on=["a"])
            df_eq(c, [], "a:int,b:int,c:int", throw=True)

        def test__join_outer(self):
            e = self.engine

            a = fa.as_fugue_engine_df(e, [], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [], "c:str,a:int")
            c = fa.left_outer_join(a, b)
            df_eq(c, [], "a:int,b:int,c:str", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.right_outer_join(a, b)
            df_eq(c, [], "a:int,b:str,c:int", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [], "c:str,a:int")
            c = fa.full_outer_join(a, b)
            df_eq(c, [], "a:int,b:str,c:str", throw=True)

            a = fa.as_fugue_engine_df(e, [[1, "2"], [3, "4"]], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [["6", 1], ["2", 7]], "c:str,a:int")
            c = fa.join(a, b, how="left_OUTER", on=["a"])
            df_eq(c, [[1, "2", "6"], [3, "4", None]], "a:int,b:str,c:str", throw=True)
            c = fa.join(b, a, how="left_outer", on=["a"])
            df_eq(c, [["6", 1, "2"], ["2", 7, None]], "c:str,a:int,b:str", throw=True)

            a = fa.as_fugue_engine_df(e, [[1, "2"], [3, "4"]], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:double,a:int")
            c = fa.join(a, b, how="left_OUTER", on=["a"])
            df_eq(
                c, [[1, "2", 6.0], [3, "4", None]], "a:int,b:str,c:double", throw=True
            )
            c = fa.join(b, a, how="left_outer", on=["a"])
            # assert c.as_pandas().values.tolist()[1][2] is None
            df_eq(
                c, [[6.0, 1, "2"], [2.0, 7, None]], "c:double,a:int,b:str", throw=True
            )

            a = fa.as_fugue_engine_df(e, [[1, "2"], [3, "4"]], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [["6", 1], ["2", 7]], "c:str,a:int")
            c = fa.join(a, b, how="right_outer", on=["a"])
            # assert c.as_pandas().values.tolist()[1][1] is None
            df_eq(c, [[1, "2", "6"], [7, None, "2"]], "a:int,b:str,c:str", throw=True)

            c = fa.join(a, b, how="full_outer", on=["a"])
            df_eq(
                c,
                [[1, "2", "6"], [3, "4", None], [7, None, "2"]],
                "a:int,b:str,c:str",
                throw=True,
            )

        def test__join_outer_pandas_incompatible(self):
            e = self.engine

            a = fa.as_fugue_engine_df(e, [[1, "2"], [3, "4"]], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.join(a, b, how="left_OUTER", on=["a"])
            df_eq(
                c,
                [[1, "2", 6], [3, "4", None]],
                "a:int,b:str,c:int",
                throw=True,
            )
            c = fa.join(b, a, how="left_outer", on=["a"])
            df_eq(c, [[6, 1, "2"], [2, 7, None]], "c:int,a:int,b:str", throw=True)

            a = fa.as_fugue_engine_df(e, [[1, "2"], [3, "4"]], "a:int,b:str")
            b = fa.as_fugue_engine_df(e, [[True, 1], [False, 7]], "c:bool,a:int")
            c = fa.join(a, b, how="left_OUTER", on=["a"])
            df_eq(c, [[1, "2", True], [3, "4", None]], "a:int,b:str,c:bool", throw=True)
            c = fa.join(b, a, how="left_outer", on=["a"])
            df_eq(
                c, [[True, 1, "2"], [False, 7, None]], "c:bool,a:int,b:str", throw=True
            )

        def test__join_semi(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.join(a, b, how="semi", on=["a"])
            df_eq(c, [[1, 2]], "a:int,b:int", throw=True)
            c = fa.semi_join(b, a)
            df_eq(c, [[6, 1]], "c:int,a:int", throw=True)

            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.join(a, b, how="semi", on=["a"])
            df_eq(c, [], "a:int,b:int", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.join(a, b, how="semi", on=["a"])
            df_eq(c, [], "a:int,b:int", throw=True)

        def test__join_anti(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.join(a, b, how="anti", on=["a"])
            df_eq(c, [[3, 4]], "a:int,b:int", throw=True)
            c = fa.anti_join(b, a)
            df_eq(c, [[2, 7]], "c:int,a:int", throw=True)

            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.join(a, b, how="anti", on=["a"])
            df_eq(c, [[1, 2], [3, 4]], "a:int,b:int", throw=True)

            a = fa.as_fugue_engine_df(e, [], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [], "c:int,a:int")
            c = fa.join(a, b, how="anti", on=["a"])
            df_eq(c, [], "a:int,b:int", throw=True)

        def test__join_with_null_keys(self):
            # SQL will not match null values
            e = self.engine
            a = fa.as_fugue_engine_df(
                e, [[1, 2, 3], [4, None, 6]], "a:double,b:double,c:int"
            )
            b = fa.as_fugue_engine_df(
                e, [[1, 2, 33], [4, None, 63]], "a:double,b:double,d:int"
            )
            c = fa.join(a, b, how="INNER")
            df_eq(c, [[1, 2, 3, 33]], "a:double,b:double,c:int,d:int", throw=True)

        def test_union(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e, [[1, 2, 3], [4, None, 6]], "a:double,b:double,c:int"
            )
            b = fa.as_fugue_engine_df(
                e, [[1, 2, 33], [4, None, 6]], "a:double,b:double,c:int"
            )
            c = fa.union(a, b)
            df_eq(
                c,
                [[1, 2, 3], [4, None, 6], [1, 2, 33]],
                "a:double,b:double,c:int",
                throw=True,
            )
            c = fa.union(a, b, distinct=False)
            df_eq(
                c,
                [[1, 2, 3], [4, None, 6], [1, 2, 33], [4, None, 6]],
                "a:double,b:double,c:int",
                throw=True,
            )
            d = fa.union(a, b, c, distinct=False)
            df_eq(
                d,
                [
                    [1, 2, 3],
                    [4, None, 6],
                    [1, 2, 33],
                    [4, None, 6],
                    [1, 2, 3],
                    [4, None, 6],
                    [1, 2, 33],
                    [4, None, 6],
                ],
                "a:double,b:double,c:int",
                throw=True,
            )

        def test_subtract(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e, [[1, 2, 3], [1, 2, 3], [4, None, 6]], "a:double,b:double,c:int"
            )
            b = fa.as_fugue_engine_df(
                e, [[1, 2, 33], [4, None, 6]], "a:double,b:double,c:int"
            )
            c = fa.subtract(a, b)
            df_eq(
                c,
                [[1, 2, 3]],
                "a:double,b:double,c:int",
                throw=True,
            )
            x = fa.as_fugue_engine_df(e, [[1, 2, 33]], "a:double,b:double,c:int")
            y = fa.as_fugue_engine_df(e, [[4, None, 6]], "a:double,b:double,c:int")
            z = fa.subtract(a, x, y)
            df_eq(
                z,
                [[1, 2, 3]],
                "a:double,b:double,c:int",
                throw=True,
            )
            # TODO: EXCEPT ALL is not implemented (QPD issue)
            # c = fa.subtract(a, b, distinct=False)
            # df_eq(
            #     c,
            #     [[1, 2, 3], [1, 2, 3]],
            #     "a:double,b:double,c:int",
            #     throw=True,
            # )

        def test_intersect(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e, [[1, 2, 3], [4, None, 6], [4, None, 6]], "a:double,b:double,c:int"
            )
            b = fa.as_fugue_engine_df(
                e,
                [[1, 2, 33], [4, None, 6], [4, None, 6], [4, None, 6]],
                "a:double,b:double,c:int",
            )
            c = fa.intersect(a, b)
            df_eq(
                c,
                [[4, None, 6]],
                "a:double,b:double,c:int",
                throw=True,
            )
            x = fa.as_fugue_engine_df(
                e,
                [[1, 2, 33]],
                "a:double,b:double,c:int",
            )
            y = fa.as_fugue_engine_df(
                e,
                [[4, None, 6], [4, None, 6], [4, None, 6]],
                "a:double,b:double,c:int",
            )
            z = fa.intersect(a, x, y)
            df_eq(
                z,
                [],
                "a:double,b:double,c:int",
                throw=True,
            )
            # TODO: INTERSECT ALL is not implemented (QPD issue)
            # c = fa.intersect(a, b, distinct=False)
            # df_eq(
            #     c,
            #     [[4, None, 6], [4, None, 6]],
            #     "a:double,b:double,c:int",
            #     throw=True,
            # )

        def test_distinct(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e, [[4, None, 6], [1, 2, 3], [4, None, 6]], "a:double,b:double,c:int"
            )
            c = fa.distinct(a)
            df_eq(
                c,
                [[4, None, 6], [1, 2, 3]],
                "a:double,b:double,c:int",
                throw=True,
            )

        def test_dropna(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e,
                [[4, None, 6], [1, 2, 3], [4, None, None]],
                "a:double,b:double,c:double",
            )
            c = fa.dropna(a)  # default
            d = fa.dropna(a, how="all")
            f = fa.dropna(a, how="any", thresh=2)
            g = fa.dropna(a, how="any", subset=["a", "c"])
            h = fa.dropna(a, how="any", thresh=1, subset=["a", "c"])
            df_eq(
                c,
                [[1, 2, 3]],
                "a:double,b:double,c:double",
                throw=True,
            )
            df_eq(
                d,
                [[4, None, 6], [1, 2, 3], [4, None, None]],
                "a:double,b:double,c:double",
                throw=True,
            )
            df_eq(
                f, [[4, None, 6], [1, 2, 3]], "a:double,b:double,c:double", throw=True
            )
            df_eq(
                g, [[4, None, 6], [1, 2, 3]], "a:double,b:double,c:double", throw=True
            )
            df_eq(
                h,
                [[4, None, 6], [1, 2, 3], [4, None, None]],
                "a:double,b:double,c:double",
                throw=True,
            )

        def test_fillna(self):
            e = self.engine
            a = fa.as_fugue_engine_df(
                e,
                [[4, None, 6], [1, 2, 3], [4, None, None]],
                "a:double,b:double,c:double",
            )
            c = fa.fillna(a, value=1)
            d = fa.fillna(a, {"b": 99, "c": -99})
            f = fa.fillna(a, value=-99, subset=["c"])
            g = fa.fillna(a, {"b": 99, "c": -99}, subset=["c"])  # subset ignored
            df_eq(
                c,
                [[4, 1, 6], [1, 2, 3], [4, 1, 1]],
                "a:double,b:double,c:double",
                throw=True,
            )
            df_eq(
                d,
                [[4, 99, 6], [1, 2, 3], [4, 99, -99]],
                "a:double,b:double,c:double",
                throw=True,
            )
            df_eq(
                f,
                [[4, None, 6], [1, 2, 3], [4, None, -99]],
                "a:double,b:double,c:double",
                throw=True,
            )
            df_eq(g, d, throw=True)
            raises(ValueError, lambda: fa.fillna(a, {"b": None, c: "99"}))
            raises(ValueError, lambda: fa.fillna(a, None))
            # raises(ValueError, lambda: fa.fillna(a, ["b"]))

        def test_sample(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[x] for x in range(100)], "a:int")

            with raises(ValueError):
                fa.sample(a)  # must set one
            with raises(ValueError):
                fa.sample(a, n=90, frac=0.9)  # can't set both

            f = fa.sample(a, frac=0.8, replace=False)
            g = fa.sample(a, frac=0.8, replace=True)
            h = fa.sample(a, frac=0.8, seed=1)
            h2 = fa.sample(a, frac=0.8, seed=1)
            i = fa.sample(a, frac=0.8, seed=2)
            assert not df_eq(f, g, throw=False)
            df_eq(h, h2, throw=True)
            assert not df_eq(h, i, throw=False)
            assert abs(len(i.as_array()) - 80) < 10

        def test_take(self):
            e = self.engine
            ps = dict(by=["a"], presort="b DESC,c DESC")
            ps2 = dict(by=["c"], presort="b ASC")
            a = fa.as_fugue_engine_df(
                e,
                [
                    ["a", 2, 3],
                    ["a", 3, 4],
                    ["b", 1, 2],
                    ["b", 2, 2],
                    [None, 4, 2],
                    [None, 2, 1],
                ],
                "a:str,b:int,c:long",
            )
            b = fa.take(a, n=1, presort="b desc")
            c = fa.take(a, n=2, presort="a desc", na_position="first")
            d = fa.take(a, n=1, presort="a asc, b desc", partition=ps)
            f = fa.take(a, n=1, presort=None, partition=ps2)
            g = fa.take(a, n=2, presort="a desc", na_position="last")
            h = fa.take(a, n=2, presort="a", na_position="first")
            df_eq(
                b,
                [[None, 4, 2]],
                "a:str,b:int,c:long",
                throw=True,
            )
            df_eq(
                c,
                [[None, 4, 2], [None, 2, 1]],
                "a:str,b:int,c:long",
                throw=True,
            )
            df_eq(
                d,
                [["a", 3, 4], ["b", 2, 2], [None, 4, 2]],
                "a:str,b:int,c:long",
                throw=True,
            )
            df_eq(
                f,
                [["a", 2, 3], ["a", 3, 4], ["b", 1, 2], [None, 2, 1]],
                "a:str,b:int,c:long",
                throw=True,
            )
            df_eq(
                g,
                [["b", 1, 2], ["b", 2, 2]],
                "a:str,b:int,c:long",
                throw=True,
            )
            df_eq(
                h,
                [
                    [None, 4, 2],
                    [None, 2, 1],
                ],
                "a:str,b:int,c:long",
                throw=True,
            )
            raises(ValueError, lambda: fa.take(a, n=0.5, presort=None))

        def test_sample_n(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[x] for x in range(100)], "a:int")

            b = fa.sample(a, n=90, replace=False)
            c = fa.sample(a, n=90, replace=True)
            d = fa.sample(a, n=90, seed=1)
            d2 = fa.sample(a, n=90, seed=1)
            e = fa.sample(a, n=90, seed=2)
            assert not df_eq(b, c, throw=False)
            df_eq(d, d2, throw=True)
            assert not df_eq(d, e, throw=False)
            assert abs(len(e.as_array()) - 90) < 2

        def test__serialize_by_partition(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4], [1, 5]], "a:int,b:int")
            s = e._serialize_by_partition(
                a, PartitionSpec(by=["a"], presort="b"), df_name="_0"
            )
            assert s.count() == 2
            s = fa.persist(e._serialize_by_partition(a, PartitionSpec(), df_name="_0"))
            assert s.count() == 1
            s = fa.persist(
                e._serialize_by_partition(a, PartitionSpec(by=["x"]), df_name="_0")
            )
            assert s.count() == 1

        def test_zip(self):
            ps = PartitionSpec(by=["a"], presort="b DESC,c DESC")
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4], [1, 5]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            sa = e._serialize_by_partition(a, ps, df_name="_0")
            sb = e._serialize_by_partition(b, ps, df_name="_1")
            # test zip with serialized dfs
            z1 = fa.persist(e.zip(sa, sb, how="inner", partition_spec=ps))
            assert 1 == z1.count()
            assert not z1.metadata.get("serialized_has_name", False)
            z2 = fa.persist(e.zip(sa, sb, how="left_outer", partition_spec=ps))
            assert 2 == z2.count()

            # can't have duplicated keys
            raises(ValueError, lambda: e.zip(sa, sa, how="inner", partition_spec=ps))
            # not support semi or anti
            raises(
                InvalidOperationError,
                lambda: e.zip(sa, sa, how="anti", partition_spec=ps),
            )
            raises(
                InvalidOperationError,
                lambda: e.zip(sa, sa, how="leftsemi", partition_spec=ps),
            )
            raises(
                InvalidOperationError,
                lambda: e.zip(sa, sa, how="LEFT SEMI", partition_spec=ps),
            )
            # can't specify keys for cross join
            raises(
                InvalidOperationError,
                lambda: e.zip(sa, sa, how="cross", partition_spec=ps),
            )

            # test zip with unserialized dfs
            z3 = fa.persist(e.zip(a, b, partition_spec=ps))
            df_eq(z1, z3, throw=True)
            z3 = fa.persist(e.zip(a, sb, partition_spec=ps))
            df_eq(z1, z3, throw=True)
            z3 = fa.persist(e.zip(sa, b, partition_spec=ps))
            df_eq(z1, z3, throw=True)

            z4 = fa.persist(e.zip(a, b, how="left_outer", partition_spec=ps))
            df_eq(z2, z4, throw=True)
            z4 = fa.persist(e.zip(a, sb, how="left_outer", partition_spec=ps))
            df_eq(z2, z4, throw=True)
            z4 = fa.persist(e.zip(sa, b, how="left_outer", partition_spec=ps))
            df_eq(z2, z4, throw=True)

            z5 = fa.persist(e.zip(a, b, how="cross"))
            assert z5.count() == 1
            assert len(z5.schema) == 2
            z6 = fa.persist(e.zip(sa, b, how="cross"))
            assert z6.count() == 2
            assert len(z6.schema) == 3

            z7 = e.zip(a, b, df1_name="x", df2_name="y")
            z7.show()
            assert z7.metadata.get("serialized_has_name", False)

        def test_zip_all(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4], [1, 5]], "a:int,b:int")
            z = fa.persist(e.zip_all(DataFrames(a)))
            assert 1 == z.count()
            assert z.metadata.get("serialized", False)
            assert not z.metadata.get("serialized_has_name", False)
            z = fa.persist(e.zip_all(DataFrames(x=a)))
            assert 1 == z.count()
            assert z.metadata.get("serialized", False)
            assert z.metadata.get("serialized_has_name", False)
            z = fa.persist(
                e.zip_all(DataFrames(x=a), partition_spec=PartitionSpec(by=["a"]))
            )
            assert 2 == z.count()
            assert z.metadata.get("serialized", False)
            assert z.metadata.get("serialized_has_name", False)

            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "d:int,a:int")
            z = fa.persist(e.zip_all(DataFrames(a, b, c)))
            assert 1 == z.count()
            assert not z.metadata.get("serialized_has_name", False)
            z = fa.persist(e.zip_all(DataFrames(x=a, y=b, z=c)))
            assert 1 == z.count()
            assert z.metadata.get("serialized_has_name", False)

            z = fa.persist(e.zip_all(DataFrames(b, b)))
            assert 2 == z.count()
            assert not z.metadata.get("serialized_has_name", False)
            assert ["a", "c"] in z.schema
            z = fa.persist(e.zip_all(DataFrames(x=b, y=b)))
            assert 2 == z.count()
            assert z.metadata.get("serialized_has_name", False)
            assert ["a", "c"] in z.schema

            z = fa.persist(
                e.zip_all(DataFrames(b, b), partition_spec=PartitionSpec(by=["a"]))
            )
            assert 2 == z.count()
            assert not z.metadata.get("serialized_has_name", False)
            assert "c" not in z.schema

        def test_comap(self):
            ps = PartitionSpec(presort="b,c")
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4], [1, 5]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            z1 = fa.persist(e.zip(a, b))
            z2 = fa.persist(e.zip(a, b, partition_spec=ps, how="left_outer"))
            z3 = fa.persist(
                e._serialize_by_partition(a, partition_spec=ps, df_name="_x")
            )
            z4 = fa.persist(e.zip(a, b, partition_spec=ps, how="cross"))

            def comap(cursor, dfs):
                assert not dfs.has_key
                v = ",".join([k + str(v.count()) for k, v in dfs.items()])
                keys = cursor.key_value_array
                if len(keys) == 0:
                    return ArrayDataFrame([[v]], "v:str")
                return ArrayDataFrame([keys + [v]], cursor.key_schema + "v:str")

            def on_init(partition_no, dfs):
                assert not dfs.has_key
                assert partition_no >= 0
                assert len(dfs) > 0

            res = e.comap(
                z1,
                comap,
                "a:int,v:str",
                PartitionSpec(),
                on_init=on_init,
            )
            df_eq(res, [[1, "_02,_11"]], "a:int,v:str", throw=True)

            # for outer joins, the NULL will be filled with empty dataframe
            res = e.comap(z2, comap, "a:int,v:str", PartitionSpec())
            df_eq(
                res,
                [[1, "_02,_11"], [3, "_01,_10"]],
                "a:int,v:str",
                throw=True,
            )

            res = e.comap(z3, comap, "v:str", PartitionSpec())
            df_eq(res, [["_03"]], "v:str", throw=True)

            res = e.comap(z4, comap, "v:str", PartitionSpec())
            df_eq(res, [["_03,_12"]], "v:str", throw=True)

        def test_comap_with_key(self):
            e = self.engine
            a = fa.as_fugue_engine_df(e, [[1, 2], [3, 4], [1, 5]], "a:int,b:int")
            b = fa.as_fugue_engine_df(e, [[6, 1], [2, 7]], "c:int,a:int")
            c = fa.as_fugue_engine_df(e, [[6, 1]], "c:int,a:int")
            z1 = fa.persist(e.zip(a, b, df1_name="x", df2_name="y"))
            z2 = fa.persist(e.zip_all(DataFrames(x=a, y=b, z=b)))
            z3 = fa.persist(
                e.zip_all(DataFrames(z=c), partition_spec=PartitionSpec(by=["a"]))
            )

            def comap(cursor, dfs):
                assert dfs.has_key
                v = ",".join([k + str(v.count()) for k, v in dfs.items()])
                keys = cursor.key_value_array
                # if len(keys) == 0:
                #    return ArrayDataFrame([[v]], "v:str")
                return ArrayDataFrame([keys + [v]], cursor.key_schema + "v:str")

            def on_init(partition_no, dfs):
                assert dfs.has_key
                assert partition_no >= 0
                assert len(dfs) > 0

            res = e.comap(
                z1,
                comap,
                "a:int,v:str",
                PartitionSpec(),
                on_init=on_init,
            )
            df_eq(res, [[1, "x2,y1"]], "a:int,v:str", throw=True)

            res = e.comap(
                z2,
                comap,
                "a:int,v:str",
                PartitionSpec(),
                on_init=on_init,
            )
            df_eq(res, [[1, "x2,y1,z1"]], "a:int,v:str", throw=True)

            res = e.comap(
                z3,
                comap,
                "a:int,v:str",
                PartitionSpec(),
                on_init=on_init,
            )
            df_eq(res, [[1, "z1"]], "a:int,v:str", throw=True)

        @pytest.fixture(autouse=True)
        def init_tmpdir(self, tmpdir):
            self.tmpdir = tmpdir

        def test_save_single_and_load_parquet(self):
            e = self.engine
            b = ArrayDataFrame([[6, 1], [2, 7]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            e.fs.makedirs(path, recreate=True)
            # over write folder with single file
            fa.save(b, path, format_hint="parquet", force_single=True)
            assert e.fs.isfile(path)
            c = fa.load(path, format_hint="parquet", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 6], [7, 2]], "a:long,c:int", throw=True)

            # overwirte single with folder (if applicable)
            b = ArrayDataFrame([[60, 1], [20, 7]], "c:int,a:long")
            fa.save(b, path, format_hint="parquet", mode="overwrite")
            c = fa.load(path, format_hint="parquet", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 60], [7, 20]], "a:long,c:int", throw=True)

        def test_save_and_load_parquet(self):
            b = ArrayDataFrame([[6, 1], [2, 7]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(b, path, format_hint="parquet")
            c = fa.load(path, format_hint="parquet", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 6], [7, 2]], "a:long,c:int", throw=True)

        def test_load_parquet_folder(self):
            native = NativeExecutionEngine()
            a = ArrayDataFrame([[6, 1]], "c:int,a:long")
            b = ArrayDataFrame([[2, 7], [4, 8]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(a, os.path.join(path, "a.parquet"), engine=native)
            fa.save(b, os.path.join(path, "b.parquet"), engine=native)
            FileSystem().touch(os.path.join(path, "_SUCCESS"))
            c = fa.load(path, format_hint="parquet", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 6], [7, 2], [8, 4]], "a:long,c:int", throw=True)

        def test_load_parquet_files(self):
            native = NativeExecutionEngine()
            a = ArrayDataFrame([[6, 1]], "c:int,a:long")
            b = ArrayDataFrame([[2, 7], [4, 8]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            f1 = os.path.join(path, "a.parquet")
            f2 = os.path.join(path, "b.parquet")
            fa.save(a, f1, engine=native)
            fa.save(b, f2, engine=native)
            c = fa.load(
                [f1, f2], format_hint="parquet", columns=["a", "c"], as_fugue=True
            )
            df_eq(c, [[1, 6], [7, 2], [8, 4]], "a:long,c:int", throw=True)

        def test_save_single_and_load_csv(self):
            e = self.engine
            b = ArrayDataFrame([[6.1, 1.1], [2.1, 7.1]], "c:double,a:double")
            path = os.path.join(self.tmpdir, "a", "b")
            e.fs.makedirs(path, recreate=True)
            # over write folder with single file
            fa.save(b, path, format_hint="csv", header=True, force_single=True)
            assert e.fs.isfile(path)
            c = fa.load(
                path, format_hint="csv", header=True, infer_schema=False, as_fugue=True
            )
            df_eq(c, [["6.1", "1.1"], ["2.1", "7.1"]], "c:str,a:str", throw=True)

            c = fa.load(
                path, format_hint="csv", header=True, infer_schema=True, as_fugue=True
            )
            df_eq(c, [[6.1, 1.1], [2.1, 7.1]], "c:double,a:double", throw=True)

            with raises(ValueError):
                c = fa.load(
                    path,
                    format_hint="csv",
                    header=True,
                    infer_schema=True,
                    columns="c:str,a:str",  # invalid to set schema when infer schema
                    as_fugue=True,
                )

            c = fa.load(
                path,
                format_hint="csv",
                header=True,
                infer_schema=False,
                columns=["a", "c"],
                as_fugue=True,
            )
            df_eq(c, [["1.1", "6.1"], ["7.1", "2.1"]], "a:str,c:str", throw=True)

            c = fa.load(
                path,
                format_hint="csv",
                header=True,
                infer_schema=False,
                columns="a:double,c:double",
                as_fugue=True,
            )
            df_eq(c, [[1.1, 6.1], [7.1, 2.1]], "a:double,c:double", throw=True)

            # overwirte single with folder (if applicable)
            b = ArrayDataFrame([[60.1, 1.1], [20.1, 7.1]], "c:double,a:double")
            fa.save(b, path, format_hint="csv", header=True, mode="overwrite")
            c = fa.load(
                path,
                format_hint="csv",
                header=True,
                infer_schema=False,
                columns=["a", "c"],
                as_fugue=True,
            )
            df_eq(c, [["1.1", "60.1"], ["7.1", "20.1"]], "a:str,c:str", throw=True)

        def test_save_single_and_load_csv_no_header(self):
            e = self.engine
            b = ArrayDataFrame([[6.1, 1.1], [2.1, 7.1]], "c:double,a:double")
            path = os.path.join(self.tmpdir, "a", "b")
            e.fs.makedirs(path, recreate=True)
            # over write folder with single file
            fa.save(b, path, format_hint="csv", header=False, force_single=True)
            assert e.fs.isfile(path)

            with raises(ValueError):
                c = fa.load(
                    path,
                    format_hint="csv",
                    header=False,
                    infer_schema=False,
                    as_fugue=True
                    # when header is False, must set columns
                )

            c = fa.load(
                path,
                format_hint="csv",
                header=False,
                infer_schema=False,
                columns=["c", "a"],
                as_fugue=True,
            )
            df_eq(c, [["6.1", "1.1"], ["2.1", "7.1"]], "c:str,a:str", throw=True)

            c = fa.load(
                path,
                format_hint="csv",
                header=False,
                infer_schema=True,
                columns=["c", "a"],
                as_fugue=True,
            )
            df_eq(c, [[6.1, 1.1], [2.1, 7.1]], "c:double,a:double", throw=True)

            with raises(ValueError):
                c = fa.load(
                    path,
                    format_hint="csv",
                    header=False,
                    infer_schema=True,
                    columns="c:double,a:double",
                    as_fugue=True,
                )

            c = fa.load(
                path,
                format_hint="csv",
                header=False,
                infer_schema=False,
                columns="c:double,a:str",
                as_fugue=True,
            )
            df_eq(c, [[6.1, "1.1"], [2.1, "7.1"]], "c:double,a:str", throw=True)

        def test_save_and_load_csv(self):
            b = ArrayDataFrame([[6.1, 1.1], [2.1, 7.1]], "c:double,a:double")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(b, path, format_hint="csv", header=True)
            c = fa.load(
                path,
                format_hint="csv",
                header=True,
                infer_schema=True,
                columns=["a", "c"],
                as_fugue=True,
            )
            df_eq(c, [[1.1, 6.1], [7.1, 2.1]], "a:double,c:double", throw=True)

        def test_load_csv_folder(self):
            native = NativeExecutionEngine()
            a = ArrayDataFrame([[6.1, 1.1]], "c:double,a:double")
            b = ArrayDataFrame([[2.1, 7.1], [4.1, 8.1]], "c:double,a:double")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(
                a,
                os.path.join(path, "a.csv"),
                format_hint="csv",
                header=True,
                engine=native,
            )
            fa.save(
                b,
                os.path.join(path, "b.csv"),
                format_hint="csv",
                header=True,
                engine=native,
            )
            FileSystem().touch(os.path.join(path, "_SUCCESS"))
            c = fa.load(
                path,
                format_hint="csv",
                header=True,
                infer_schema=True,
                columns=["a", "c"],
                as_fugue=True,
            )
            df_eq(
                c, [[1.1, 6.1], [7.1, 2.1], [8.1, 4.1]], "a:double,c:double", throw=True
            )

        def test_save_single_and_load_json(self):
            e = self.engine
            b = ArrayDataFrame([[6, 1], [2, 7]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            e.fs.makedirs(path, recreate=True)
            # over write folder with single file
            fa.save(b, path, format_hint="json", force_single=True)
            assert e.fs.isfile(path)
            c = fa.load(path, format_hint="json", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 6], [7, 2]], "a:long,c:long", throw=True)

            # overwirte single with folder (if applicable)
            b = ArrayDataFrame([[60, 1], [20, 7]], "c:long,a:long")
            fa.save(b, path, format_hint="json", mode="overwrite")
            c = fa.load(path, format_hint="json", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 60], [7, 20]], "a:long,c:long", throw=True)

        def test_save_and_load_json(self):
            e = self.engine
            b = ArrayDataFrame([[6, 1], [3, 4], [2, 7], [4, 8], [6, 7]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(
                e.repartition(fa.as_fugue_engine_df(e, b), PartitionSpec(num=2)),
                path,
                format_hint="json",
            )
            c = fa.load(path, format_hint="json", columns=["a", "c"], as_fugue=True)
            df_eq(
                c, [[1, 6], [7, 2], [4, 3], [8, 4], [7, 6]], "a:long,c:long", throw=True
            )

        def test_load_json_folder(self):
            native = NativeExecutionEngine()
            a = ArrayDataFrame([[6, 1], [3, 4]], "c:int,a:long")
            b = ArrayDataFrame([[2, 7], [4, 8]], "c:int,a:long")
            path = os.path.join(self.tmpdir, "a", "b")
            fa.save(a, os.path.join(path, "a.json"), format_hint="json", engine=native)
            fa.save(b, os.path.join(path, "b.json"), format_hint="json", engine=native)
            FileSystem().touch(os.path.join(path, "_SUCCESS"))
            c = fa.load(path, format_hint="json", columns=["a", "c"], as_fugue=True)
            df_eq(c, [[1, 6], [7, 2], [8, 4], [4, 3]], "a:long,c:long", throw=True)

        def test_engine_api(self):
            # complimentary tests not covered by the other tests
            with fa.engine_context(self.engine):
                df1 = fa.as_fugue_df([[0, 1], [2, 3]], schema="a:long,b:long")
                df1 = fa.repartition(df1, {"num": 2})
                df1 = fa.get_native_as_df(fa.broadcast(df1))
                df2 = pd.DataFrame([[0, 1], [2, 3]], columns=["a", "b"])
                df3 = fa.union(df1, df2, as_fugue=False)
                assert fa.is_df(df3) and not isinstance(df3, DataFrame)
                df4 = fa.union(df1, df2, as_fugue=True)
                assert isinstance(df4, DataFrame)
                df_eq(df4, fa.as_pandas(df3), throw=True)


def select_top(cursor, data):
    return ArrayDataFrame([cursor.row], cursor.row_schema)


class BinaryObject(object):
    def __init__(self, data=None):
        self.data = data


def binary_map(cursor, df):
    arr = df.as_array(type_safe=True)
    for i in range(len(arr)):
        obj = pickle.loads(arr[i][0])
        obj.data += "x"
        arr[i][0] = pickle.dumps(obj)
    return ArrayDataFrame(arr, df.schema)



import re
from typing import Any, Dict, Optional

from triad import assert_or_throw

from ..collections.yielded import Yielded
from ..exceptions import FugueSQLError
from ..workflow.workflow import FugueWorkflow, WorkflowDataFrame

MATCH_QUOTED_STRING = r"([\"'])(({|%|})*)\1"


def fill_sql_template(sql: str, params: Dict[str, Any]):
    """Prepare string to be executed, inserts params into sql template
    ---
    :param sql: jinja compatible template
    :param params: params to be inserted into template
    """
    import jinja2
    from jinja2 import Template

    try:
        if "self" in params:
            params = {k: v for k, v in params.items() if k != "self"}
        single_quote_pattern = "'{{% raw %}}{}{{% endraw %}}'"
        double_quote_pattern = '"{{% raw %}}{}{{% endraw %}}"'
        new_sql = re.sub(
            MATCH_QUOTED_STRING,
            lambda pattern: double_quote_pattern.format(pattern.group(2))
            if pattern.group(1) == '"'
            else single_quote_pattern.format(pattern.group(2)),
            sql,
        )

        template = Template(new_sql)

    except jinja2.exceptions.TemplateSyntaxError:

        template = Template(sql)

    return template.render(**params)


class LazyWorkflowDataFrame:
    def __init__(self, key: str, df: Any, workflow: FugueWorkflow):
        self._key = key
        self._df = df
        self._workflow = workflow
        self._wdf: Optional[WorkflowDataFrame] = None

    def get_df(self) -> WorkflowDataFrame:
        if self._wdf is None:
            self._wdf = self._get_df()
        return self._wdf

    def _get_df(self) -> WorkflowDataFrame:
        if isinstance(self._df, Yielded):
            return self._workflow.df(self._df)
        if isinstance(self._df, WorkflowDataFrame):
            assert_or_throw(
                self._df.workflow is self._workflow,
                lambda: FugueSQLError(
                    f"{self._key}, {self._df} is from another workflow"
                ),
            )
            return self._df
        return self._workflow.df(self._df)



# flake8: noqa

from ._workflow_context import FugueWorkflowContext
from .api import *
from .input import register_raw_df_type
from .module import module
from .workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames



# flake8: noqa
from fugue.extensions._builtins.creators import Load, CreateData
from fugue.extensions._builtins.outputters import (
    AssertEqual,
    AssertNotEqual,
    RunOutputTransformer,
    Save,
    Show,
)
from fugue.extensions._builtins.processors import (
    Aggregate,
    AlterColumns,
    Assign,
    Distinct,
    DropColumns,
    Dropna,
    Fillna,
    Filter,
    Rename,
    RunJoin,
    RunSetOperation,
    RunSQLSelect,
    RunTransformer,
    Sample,
    SaveAndUse,
    Select,
    SelectColumns,
    Take,
    Zip,
)



# pylint: disable-all

from datetime import date, datetime
from typing import Any
from unittest import TestCase

import numpy as np
import pandas as pd
from pytest import raises

import fugue.api as fi
from fugue.dataframe import ArrowDataFrame, DataFrame
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.exceptions import FugueDataFrameOperationError, FugueDatasetEmptyError


class DataFrameTests(object):
    """DataFrame level general test suite.
    All new DataFrame types should pass this test suite.
    """

    class Tests(TestCase):
        @classmethod
        def setUpClass(cls):
            pass

        @classmethod
        def tearDownClass(cls):
            pass

        def df(self, data: Any = None, schema: Any = None) -> Any:  # pragma: no cover
            raise NotImplementedError

        def test_native(self):
            df = self.df([1], "a:int")
            assert fi.is_df(df)
            fdf = fi.as_fugue_df(df)
            assert isinstance(fdf, DataFrame)
            assert fi.is_df(fdf)
            ndf = fi.get_native_as_df(fdf)
            assert fi.is_df(ndf)
            assert not isinstance(ndf, DataFrame)
            ndf2 = fi.get_native_as_df(ndf)
            assert ndf2 is ndf

        def test_peek(self):
            df = self.df([], "x:str,y:double")
            raises(FugueDatasetEmptyError, lambda: fi.peek_array(df))
            raises(FugueDatasetEmptyError, lambda: fi.peek_dict(df))

            df = self.df([["a", 1.0], ["b", 2.0]], "x:str,y:double")
            assert not fi.is_bounded(df) or 2 == fi.count(df)
            assert not fi.is_empty(df)
            assert ["a", 1.0] == fi.peek_array(df)
            assert dict(x="a", y=1.0) == fi.peek_dict(df)

        def test_as_pandas(self):
            df = self.df([["a", 1.0], ["b", 2.0]], "x:str,y:double")
            pdf = fi.as_pandas(df)
            assert [["a", 1.0], ["b", 2.0]] == pdf.values.tolist()

            df = self.df([], "x:str,y:double")
            pdf = fi.as_pandas(df)
            assert [] == pdf.values.tolist()
            assert fi.is_local(pdf)

        def test_as_local(self):
            with raises(NotImplementedError):
                fi.as_local(10)
            with raises(NotImplementedError):
                fi.as_local_bounded(10)

            df = self.df([["a", 1.0], ["b", 2.0]], "x:str,y:double")
            ldf = fi.as_local(df)
            assert fi.is_local(ldf)
            lbdf = fi.as_local_bounded(df)
            assert fi.is_local(lbdf) and fi.is_bounded(lbdf)

            fdf = fi.as_fugue_df(df)
            fdf.reset_metadata({"a": 1})
            ldf = fi.as_local(fdf)
            assert ldf.metadata == {"a": 1}
            lbdf = fi.as_local_bounded(fdf)
            assert fi.is_local(lbdf) and fi.is_bounded(lbdf)
            assert ldf.metadata == {"a": 1}

        def test_drop_columns(self):
            df = fi.drop_columns(self.df([], "a:str,b:int"), ["a"])
            assert fi.get_schema(df) == "b:int"
            raises(
                FugueDataFrameOperationError, lambda: fi.drop_columns(df, ["b"])
            )  # can't be empty
            raises(
                FugueDataFrameOperationError, lambda: fi.drop_columns(df, ["x"])
            )  # cols must exist

            df = fi.drop_columns(self.df([["a", 1]], "a:str,b:int"), ["a"])
            assert fi.get_schema(df) == "b:int"
            raises(
                FugueDataFrameOperationError, lambda: fi.drop_columns(df, ["b"])
            )  # can't be empty
            raises(
                FugueDataFrameOperationError, lambda: fi.drop_columns(df, ["x"])
            )  # cols must exist
            assert [[1]] == fi.as_array(df, type_safe=True)

        def test_select(self):
            df = fi.select_columns(self.df([], "a:str,b:int"), ["b"])
            assert fi.get_schema(df) == "b:int"
            assert fi.get_column_names(df) == ["b"]
            raises(
                FugueDataFrameOperationError, lambda: fi.select_columns(df, [])
            )  # select empty
            raises(
                FugueDataFrameOperationError, lambda: fi.select_columns(df, ["a"])
            )  # not existed
            raises(
                FugueDataFrameOperationError, lambda: fi.select_columns(df, ["a"])
            )  # empty

            df = fi.select_columns(self.df([["a", 1]], "a:str,b:int"), ["b"])
            assert fi.get_schema(df) == "b:int"
            raises(
                FugueDataFrameOperationError, lambda: fi.select_columns(df, ["a"])
            )  # not existed
            raises(
                FugueDataFrameOperationError, lambda: fi.select_columns(df, ["a"])
            )  # empty
            assert [[1]] == fi.as_array(df, type_safe=True)

            df = self.df([["a", 1, 2]], "a:str,b:int,c:int")
            df_eq(
                fi.as_fugue_df(fi.select_columns(df, ["c", "a"])),
                [[2, "a"]],
                "a:str,c:int",
            )

        def test_rename(self):
            for data in [[["a", 1]], []]:
                df = self.df(data, "a:str,b:int")
                df2 = fi.rename(df, columns=dict(a="aa"))
                assert fi.get_schema(df) == "a:str,b:int"
                df_eq(fi.as_fugue_df(df2), data, "aa:str,b:int", throw=True)

            for data in [[["a", 1]], []]:
                df = self.df(data, "a:str,b:int")
                df3 = fi.rename(df, columns={})
                assert fi.get_schema(df3) == "a:str,b:int"
                df_eq(fi.as_fugue_df(df3), data, "a:str,b:int", throw=True)

        def test_rename_invalid(self):
            df = self.df([["a", 1]], "a:str,b:int")
            raises(
                FugueDataFrameOperationError,
                lambda: fi.rename(df, columns=dict(aa="ab")),
            )

        def test_as_array(self):
            for func in [
                lambda df, *args, **kwargs: fi.as_array(
                    df, *args, **kwargs, type_safe=True
                ),
                lambda df, *args, **kwargs: list(
                    fi.as_array_iterable(df, *args, **kwargs, type_safe=True)
                ),
            ]:
                df = self.df([], "a:str,b:int")
                assert [] == func(df)

                df = self.df([["a", 1]], "a:str,b:int")
                assert [["a", 1]] == func(df)
                df = self.df([["a", 1]], "a:str,b:int")
                assert [["a", 1]] == func(df, ["a", "b"])
                df = self.df([["a", 1]], "a:str,b:int")
                assert [[1, "a"]] == func(df, ["b", "a"])

                for v in [1.0, np.float64(1.0)]:
                    df = self.df([[v, 1]], "a:double,b:int")
                    d = func(df)
                    assert [[1.0, 1]] == d
                    assert isinstance(d[0][0], float)
                    assert isinstance(d[0][1], int)

        def test_as_array_special_values(self):
            for func in [
                lambda df, *args, **kwargs: fi.as_array(
                    df, *args, **kwargs, type_safe=True
                ),
                lambda df, *args, **kwargs: list(
                    fi.as_array_iterable(df, *args, **kwargs, type_safe=True)
                ),
            ]:
                df = self.df([[pd.Timestamp("2020-01-01"), 1]], "a:datetime,b:int")
                data = func(df)
                assert [[datetime(2020, 1, 1), 1]] == data
                assert isinstance(data[0][0], datetime)
                assert isinstance(data[0][1], int)

                df = self.df([[pd.NaT, 1]], "a:datetime,b:int")
                assert [[None, 1]] == func(df)

                df = self.df([[float("nan"), 1]], "a:double,b:int")
                assert [[None, 1]] == func(df)

                df = self.df([[float("inf"), 1]], "a:double,b:int")
                assert [[float("inf"), 1]] == func(df)

        def test_as_dict_iterable(self):
            df = self.df([[pd.NaT, 1]], "a:datetime,b:int")
            assert [dict(a=None, b=1)] == list(fi.as_dict_iterable(df))
            df = self.df([[pd.Timestamp("2020-01-01"), 1]], "a:datetime,b:int")
            assert [dict(a=datetime(2020, 1, 1), b=1)] == list(fi.as_dict_iterable(df))

        def test_list_type(self):
            data = [[[30, 40]]]
            df = self.df(data, "a:[int]")
            a = fi.as_array(df, type_safe=True)
            assert data == a

        def test_struct_type(self):
            data = [[{"a": 1}], [{"a": 2}]]
            df = self.df(data, "x:{a:int}")
            a = fi.as_array(df, type_safe=True)
            assert data == a

        def test_map_type(self):
            data = [[[("a", 1), ("b", 3)]], [[("b", 2)]]]
            df = self.df(data, "x:<str,int>")
            a = fi.as_array(df, type_safe=True)
            assert data == a

        def test_deep_nested_types(self):
            data = [[dict(a="1", b=[3, 4], d=1.0)], [dict(b=[30, 40])]]
            df = self.df(data, "a:{a:str,b:[int]}")
            a = fi.as_array(df, type_safe=True)
            assert [[dict(a="1", b=[3, 4])], [dict(a=None, b=[30, 40])]] == a

            data = [[[dict(b=[30, 40])]]]
            df = self.df(data, "a:[{a:str,b:[int]}]")
            a = fi.as_array(df, type_safe=True)
            assert [[[dict(a=None, b=[30, 40])]]] == a

        def test_binary_type(self):
            data = [[b"\x01\x05"]]
            df = self.df(data, "a:bytes")
            a = fi.as_array(df, type_safe=True)
            assert data == a

        def test_as_arrow(self):
            # empty
            df = self.df([], "a:int,b:int")
            assert [] == list(ArrowDataFrame(fi.as_arrow(df)).as_dict_iterable())
            assert fi.is_local(fi.as_arrow(df))
            # pd.Nat
            df = self.df([[pd.NaT, 1]], "a:datetime,b:int")
            assert [dict(a=None, b=1)] == list(
                ArrowDataFrame(fi.as_arrow(df)).as_dict_iterable()
            )
            # pandas timestamps
            df = self.df([[pd.Timestamp("2020-01-01"), 1]], "a:datetime,b:int")
            assert [dict(a=datetime(2020, 1, 1), b=1)] == list(
                ArrowDataFrame(fi.as_arrow(df)).as_dict_iterable()
            )
            # float nan, list
            data = [[[float("nan"), 2.0]]]
            df = self.df(data, "a:[float]")
            assert [[[None, 2.0]]] == ArrowDataFrame(fi.as_arrow(df)).as_array()
            # dict
            data = [[dict(b=True)]]
            df = self.df(data, "a:{b:bool}")
            assert data == ArrowDataFrame(fi.as_arrow(df)).as_array()
            # list[dict]
            data = [[[dict(b=[30, 40])]]]
            df = self.df(data, "a:[{b:[int]}]")
            assert data == ArrowDataFrame(fi.as_arrow(df)).as_array()

        def test_head(self):
            df = self.df([], "a:str,b:int")
            assert [] == fi.as_array(fi.head(df, 1))
            assert [] == fi.as_array(fi.head(df, 1, ["b"]))
            df = self.df([["a", 1]], "a:str,b:int")
            if fi.is_bounded(df):
                assert [["a", 1]] == fi.as_array(fi.head(df, 1))
            assert [[1, "a"]] == fi.as_array(fi.head(df, 1, ["b", "a"]))
            assert [] == fi.as_array(fi.head(df, 0))

            df = self.df([[0, 1], [0, 2], [1, 1], [1, 3]], "a:int,b:int")
            assert 2 == fi.count(fi.head(df, 2))
            df = self.df([[0, 1], [0, 2], [1, 1], [1, 3]], "a:int,b:int")
            assert 4 == fi.count(fi.head(df, 10))
            h = fi.head(df, 10)
            assert fi.is_local(h) and fi.is_bounded(h)

        def test_show(self):
            df = self.df([["a", 1]], "a:str,b:int")
            fi.show(df)

        def test_get_altered_schema(self):
            df = self.df([["a", 1]], "a:str,b:int")
            assert df._get_altered_schema("") == "a:str,b:int"
            assert df._get_altered_schema(None) == "a:str,b:int"
            assert df._get_altered_schema("b:str,a:str") == "a:str,b:str"
            with raises(FugueDataFrameOperationError):
                df._get_altered_schema("bb:str,a:str")
            with raises(NotImplementedError):
                df._get_altered_schema("b:binary")
            with raises(NotImplementedError):
                df._get_altered_schema("b:[str]")
            with raises(NotImplementedError):
                df._get_altered_schema("b:{x:str}")

        def test_alter_columns(self):
            # empty
            df = self.df([], "a:str,b:int")
            ndf = fi.alter_columns(df, "a:str,b:str")
            assert [] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:str,b:str"

            # no change
            df = self.df([["a", 1], ["c", None]], "a:str,b:int")
            ndf = fi.alter_columns(df, "b:int,a:str", as_fugue=True)
            assert [["a", 1], ["c", None]] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:str,b:int"

            # bool -> str
            df = self.df([["a", True], ["b", False], ["c", None]], "a:str,b:bool")
            ndf = fi.alter_columns(df, "b:str", as_fugue=True)
            actual = fi.as_array(ndf, type_safe=True)
            # Capitalization doesn't matter
            # and dataframes don't need to be consistent on capitalization
            expected1 = [["a", "True"], ["b", "False"], ["c", None]]
            expected2 = [["a", "true"], ["b", "false"], ["c", None]]
            assert expected1 == actual or expected2 == actual
            assert fi.get_schema(ndf) == "a:str,b:str"

            # int -> str
            df = self.df([["a", 1], ["c", None]], "a:str,b:int")
            ndf = fi.alter_columns(df, "b:str", as_fugue=True)
            arr = fi.as_array(ndf, type_safe=True)
            assert [["a", "1"], ["c", None]] == arr or [
                ["a", "1.0"],
                ["c", None],
            ] == arr  # in pandas case, it can't treat [1, None] as an int col
            assert fi.get_schema(ndf) == "a:str,b:str"

            # int -> double
            df = self.df([["a", 1], ["c", None]], "a:str,b:int")
            ndf = fi.alter_columns(df, "b:double", as_fugue=True)
            assert [["a", 1], ["c", None]] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:str,b:double"

            # double -> str
            df = self.df([["a", 1.1], ["b", None]], "a:str,b:double")
            data = fi.as_array(
                fi.alter_columns(df, "b:str", as_fugue=True), type_safe=True
            )
            assert [["a", "1.1"], ["b", None]] == data

            # double -> int
            df = self.df([["a", 1.0], ["b", None]], "a:str,b:double")
            data = fi.as_array(
                fi.alter_columns(df, "b:int", as_fugue=True), type_safe=True
            )
            assert [["a", 1], ["b", None]] == data

            # date -> str
            df = self.df(
                [["a", date(2020, 1, 1)], ["b", date(2020, 1, 2)], ["c", None]],
                "a:str,b:date",
            )
            data = fi.as_array(
                fi.alter_columns(df, "b:str", as_fugue=True), type_safe=True
            )
            assert [["a", "2020-01-01"], ["b", "2020-01-02"], ["c", None]] == data

            # datetime -> str
            df = self.df(
                [
                    ["a", datetime(2020, 1, 1, 3, 4, 5)],
                    ["b", datetime(2020, 1, 2, 16, 7, 8)],
                    ["c", None],
                ],
                "a:str,b:datetime",
            )
            data = fi.as_array(
                fi.alter_columns(df, "b:str", as_fugue=True), type_safe=True
            )
            assert [
                ["a", "2020-01-01 03:04:05"],
                ["b", "2020-01-02 16:07:08"],
                ["c", None],
            ] == data

            # str -> bool
            df = self.df([["a", "trUe"], ["b", "False"], ["c", None]], "a:str,b:str")
            ndf = fi.alter_columns(df, "b:bool,a:str", as_fugue=True)
            assert [["a", True], ["b", False], ["c", None]] == fi.as_array(
                ndf, type_safe=True
            )
            assert fi.get_schema(ndf) == "a:str,b:bool"

            # str -> int
            df = self.df([["a", "1"]], "a:str,b:str")
            ndf = fi.alter_columns(df, "b:int,a:str")
            assert [["a", 1]] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:str,b:int"

            # str -> double
            df = self.df([["a", "1.1"], ["b", "2"], ["c", None]], "a:str,b:str")
            ndf = fi.alter_columns(df, "b:double", as_fugue=True)
            assert [["a", 1.1], ["b", 2.0], ["c", None]] == fi.as_array(
                ndf, type_safe=True
            )
            assert fi.get_schema(ndf) == "a:str,b:double"

            # str -> date
            df = self.df(
                [["1", "2020-01-01"], ["2", "2020-01-02"], ["3", None]],
                "a:str,b:str",
            )
            ndf = fi.alter_columns(df, "b:date,a:int", as_fugue=True)
            assert [
                [1, date(2020, 1, 1)],
                [2, date(2020, 1, 2)],
                [3, None],
            ] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:int,b:date"

            # str -> datetime
            df = self.df(
                [
                    ["1", "2020-01-01 01:02:03"],
                    ["2", "2020-01-02 01:02:03"],
                    ["3", None],
                ],
                "a:str,b:str",
            )
            ndf = fi.alter_columns(df, "b:datetime,a:int", as_fugue=True)
            assert [
                [1, datetime(2020, 1, 1, 1, 2, 3)],
                [2, datetime(2020, 1, 2, 1, 2, 3)],
                [3, None],
            ] == fi.as_array(ndf, type_safe=True)
            assert fi.get_schema(ndf) == "a:int,b:datetime"

        def test_alter_columns_invalid(self):
            # invalid conversion
            with raises(Exception):
                df = self.df(
                    [["1", "x"], ["2", "y"], ["3", None]],
                    "a:str,b:str",
                )
                ndf = fi.alter_columns(df, "b:int")
                fi.show(ndf)  # lazy dataframes will force to materialize

    class NativeTests(Tests):
        def to_native_df(self, pdf: pd.DataFrame) -> Any:  # pragma: no cover
            raise NotImplementedError

        def test_get_altered_schema(self):
            pass

        def test_get_column_names(self):
            df = self.to_native_df(pd.DataFrame([[0, 1, 2]], columns=["0", "1", "2"]))
            assert fi.get_column_names(df) == ["0", "1", "2"]

        def test_rename_any_names(self):
            pdf = self.to_native_df(pd.DataFrame([[0, 1, 2]], columns=["a", "b", "c"]))
            df = fi.rename(pdf, {})
            assert fi.get_column_names(df) == ["a", "b", "c"]

            pdf = self.to_native_df(pd.DataFrame([[0, 1, 2]], columns=["0", "1", "2"]))
            df = fi.rename(pdf, {"0": "_0", "1": "_1", "2": "_2"})
            assert fi.get_column_names(df) == ["_0", "_1", "_2"]



# flake8: noqa
from fugue.extensions.outputter.convert import (
    _to_outputter,
    outputter,
    parse_outputter,
    register_outputter,
)
from fugue.extensions.outputter.outputter import Outputter



from typing import Any, Callable, Dict, Optional, List

import ibis
import ibis.expr.datatypes as dt
import pyarrow as pa
from triad import Schema, extensible_class
from triad.utils.pyarrow import TRIAD_DEFAULT_TIMESTAMP

_SPECIAL_METHODS = [
    "__getitem__",
    "__add__",
    "__bool__",
    "__div__",
    "__eq__",
    "__floordiv__",
    "__ge__",
    "__gt__",
    "__hash__",
    "__le__",
    "__lt__",
    "__mod__",
    "__mul__",
    "__ne__",
    "__neg__",
    "__pow__",
    "__radd__",
    "__rdiv__",
    "__rfloordiv__",
    "__rmod__",
    "__rmul__",
    "__rpow__",
    "__rsub__",
    "__rtruediv__",
    "__sub__",
    "__truediv__",
]

_IBIS_TO_PYARROW: Dict[dt.DataType, pa.DataType] = {
    dt.boolean: pa.bool_(),
    dt.int8: pa.int8(),
    dt.uint8: pa.uint8(),
    dt.int16: pa.int16(),
    dt.uint16: pa.uint16(),
    dt.int32: pa.int32(),
    dt.uint32: pa.uint32(),
    dt.int64: pa.int64(),
    dt.uint64: pa.uint64(),
    dt.float32: pa.float32(),
    dt.float64: pa.float64(),
    dt.string: pa.string(),
    dt.binary: pa.binary(),
    dt.date: pa.date32(),
}

_PYARROW_TO_IBIS: Dict[pa.DataType, dt.DataType] = {
    v: k for k, v in _IBIS_TO_PYARROW.items()
}


def to_ibis_schema(schema: Schema) -> ibis.Schema:
    fields = [(f.name, pa_to_ibis_type(f.type)) for f in schema.fields]
    return ibis.schema(fields)


def to_schema(
    schema: ibis.Schema,
    on_incompatible: Optional[Callable[[str, dt.DataType], pa.DataType]] = None,
) -> Schema:
    fields: List[Any] = []
    for n, t in zip(schema.names, schema.types):
        try:
            fields.append((n, ibis_to_pa_type(t)))
        except NotImplementedError:
            if on_incompatible is None:
                raise
            fields.append((n, on_incompatible(n, t)))
    return Schema(fields)


def ibis_to_pa_type(tp: dt.DataType) -> pa.DataType:
    if tp in _IBIS_TO_PYARROW:
        return _IBIS_TO_PYARROW[tp]
    if isinstance(tp, dt.Timestamp):
        if tp.timezone is None:
            return TRIAD_DEFAULT_TIMESTAMP
        else:
            return pa.timestamp("us", tp.timezone)
    if isinstance(tp, dt.Decimal) and tp.precision is not None:
        return pa.decimal128(tp.precision, 0 if tp.scale is None else tp.scale)
    if isinstance(tp, dt.Array):
        ttp = ibis_to_pa_type(tp.value_type)
        return pa.list_(ttp)
    if isinstance(tp, dt.Struct):
        fields = [pa.field(n, ibis_to_pa_type(t)) for n, t in zip(tp.names, tp.types)]
        return pa.struct(fields)
    if isinstance(tp, dt.Map):
        return pa.map_(ibis_to_pa_type(tp.key_type), ibis_to_pa_type(tp.value_type))
    raise NotImplementedError(tp)


def pa_to_ibis_type(tp: pa.DataType) -> dt.DataType:
    if tp in _PYARROW_TO_IBIS:
        return _PYARROW_TO_IBIS[tp]
    if pa.types.is_timestamp(tp):
        if tp.tz is None:
            return dt.Timestamp()
        return dt.Timestamp(timezone=str(tp.tz))
    if pa.types.is_decimal(tp):
        return dt.Decimal(tp.precision, tp.scale)
    if pa.types.is_list(tp):
        ttp = pa_to_ibis_type(tp.value_type)
        return dt.Array(value_type=ttp)
    if pa.types.is_struct(tp):
        fields = [(f.name, pa_to_ibis_type(f.type)) for f in tp]
        return dt.Struct.from_tuples(fields)
    if pa.types.is_map(tp):
        return dt.Map(
            key_type=pa_to_ibis_type(tp.key_type),
            value_type=pa_to_ibis_type(tp.item_type),
        )
    raise NotImplementedError(tp)  # pragma: no cover


def materialize(obj: "LazyIbisObject", materialize_func: Callable) -> Any:
    ctx = {k: materialize_func(v) for k, v in obj._super_lazy_internal_ctx.items()}
    return _materialize(obj, ctx)


@extensible_class
class LazyIbisObject:
    def __init__(self, obj: Any = None):
        self._super_lazy_internal_ctx: Dict[int, Any] = {}
        if obj is not None:
            self._super_lazy_internal_ctx[id(self)] = obj

    def __getattr__(self, name: str) -> Any:
        if not name.startswith("_"):
            return LazyIbisAttr(self, name)


class LazyIbisAttr(LazyIbisObject):
    def __init__(self, parent: LazyIbisObject, name: str):
        super().__init__()
        self._super_lazy_internal_ctx.update(parent._super_lazy_internal_ctx)
        self._super_lazy_internal_objs: Dict[str, Any] = dict(parent=parent, name=name)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return LazyIbisFunction(
            self._super_lazy_internal_objs["parent"],
            self._super_lazy_internal_objs["name"],
            *args,
            **kwargs
        )


class LazyIbisFunction(LazyIbisObject):
    def __init__(self, obj: LazyIbisObject, func: str, *args: Any, **kwargs: Any):
        super().__init__()
        self._super_lazy_internal_ctx.update(obj._super_lazy_internal_ctx)
        for x in args:
            if isinstance(x, LazyIbisObject):
                self._super_lazy_internal_ctx.update(x._super_lazy_internal_ctx)
        for x in kwargs.values():
            if isinstance(x, LazyIbisObject):
                self._super_lazy_internal_ctx.update(x._super_lazy_internal_ctx)
        self._super_lazy_internal_objs: Dict[str, Any] = dict(
            obj=obj, func=func, args=args, kwargs=kwargs
        )


def _wrapper(func: str):
    return lambda obj, *args, **kwargs: LazyIbisFunction(obj, func, *args, **kwargs)


for _method in _SPECIAL_METHODS:
    setattr(
        LazyIbisObject,
        _method,
        _wrapper(_method),
    )


def _materialize(obj: Any, context: Dict[int, Any]) -> Any:
    if id(obj) in context:
        return context[id(obj)]
    if isinstance(obj, list):
        v: Any = [_materialize(x, context) for x in obj]
    elif isinstance(obj, tuple):
        v = tuple(_materialize(x, context) for x in obj)
    elif isinstance(obj, dict):
        v = {k: _materialize(v, context) for k, v in obj.items()}
    elif isinstance(obj, LazyIbisFunction):
        f = getattr(
            _materialize(obj._super_lazy_internal_objs["obj"], context),
            obj._super_lazy_internal_objs["func"],
        )
        v = f(
            *_materialize(obj._super_lazy_internal_objs["args"], context),
            **_materialize(obj._super_lazy_internal_objs["kwargs"], context)
        )
    elif isinstance(obj, LazyIbisAttr):
        v = getattr(
            _materialize(obj._super_lazy_internal_objs["parent"], context),
            obj._super_lazy_internal_objs["name"],
        )
    elif isinstance(obj, LazyIbisObject):
        raise NotImplementedError  # pragma: no cover
    else:
        v = obj
    context[id(obj)] = v
    return v



import json
from abc import abstractmethod
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import pandas as pd
import pyarrow as pa
from triad import SerializableRLock
from triad.collections.schema import Schema
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw
from triad.utils.pandas_like import PD_UTILS

from .._utils.display import PrettyTable
from ..collections.yielded import Yielded
from ..dataset import (
    Dataset,
    DatasetDisplay,
    as_local,
    as_local_bounded,
    get_dataset_display,
    as_fugue_dataset,
)
from ..exceptions import FugueDataFrameOperationError

AnyDataFrame = TypeVar("AnyDataFrame", "DataFrame", object)


class DataFrame(Dataset):
    """Base class of Fugue DataFrame. Please read
    |DataFrameTutorial| to understand the concept

    :param schema: |SchemaLikeObject|

    .. note::

        This is an abstract class, and normally you don't construct it by yourself
        unless you are
        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    def __init__(self, schema: Any = None):
        super().__init__()
        if not callable(schema):
            schema = _input_schema(schema).assert_not_empty()
            schema.set_readonly()
            self._schema: Union[Schema, Callable[[], Schema]] = schema
            self._schema_discovered = True
        else:
            self._schema: Union[Schema, Callable[[], Schema]] = schema  # type: ignore
            self._schema_discovered = False
        self._lazy_schema_lock = SerializableRLock()

    @property
    def schema(self) -> Schema:
        """The schema of the dataframe"""
        if self.schema_discovered:
            # we must keep it simple because it could be called on every row by a user
            assert isinstance(self._schema, Schema)
            return self._schema  # type: ignore
        with self._lazy_schema_lock:
            self._schema = _input_schema(
                self._schema()
            ).assert_not_empty()  # type: ignore
            self._schema.set_readonly()
            self._schema_discovered = True
            return self._schema

    @property
    def schema_discovered(self) -> Schema:
        """Whether the schema has been discovered or still a lambda"""
        return self._schema_discovered

    @property
    def columns(self) -> List[str]:
        """The column names of the dataframe"""
        return self.schema.names

    @abstractmethod
    def native_as_df(self) -> AnyDataFrame:  # pragma: no cover
        """The dataframe form of the native object this Dataset class wraps.
        Dataframe form means the object contains schema information. For example
        the native an ArrayDataFrame is a python array, it doesn't contain schema
        information, and its ``native_as_df`` should be either a pandas dataframe
        or an arrow dataframe.
        """
        raise NotImplementedError

    def as_local(self) -> "LocalDataFrame":  # pragma: no cover
        """Convert this dataframe to a :class:`.LocalDataFrame`"""
        return self.as_local_bounded()

    @abstractmethod
    def as_local_bounded(self) -> "LocalBoundedDataFrame":  # pragma: no cover
        """Convert this dataframe to a :class:`.LocalBoundedDataFrame`"""
        raise NotImplementedError

    @abstractmethod
    def peek_array(self) -> List[Any]:  # pragma: no cover
        """Peek the first row of the dataframe as array

        :raises FugueDatasetEmptyError: if it is empty
        """
        raise NotImplementedError

    def peek_dict(self) -> Dict[str, Any]:
        """Peek the first row of the dataframe as dict

        :raises FugueDatasetEmptyError: if it is empty
        """
        arr = self.peek_array()
        return {self.columns[i]: arr[i] for i in range(len(self.columns))}

    def as_pandas(self) -> pd.DataFrame:
        """Convert to pandas DataFrame"""
        pdf = pd.DataFrame(self.as_array(), columns=self.columns)
        if len(pdf) == 0:  # TODO: move to triad
            return pd.DataFrame(
                {
                    k: pd.Series(dtype=v.type.to_pandas_dtype())
                    for k, v in self.schema.items()
                }
            )
        return PD_UTILS.enforce_type(pdf, self.schema.pa_schema, null_safe=True)

    def as_arrow(self, type_safe: bool = False) -> pa.Table:
        """Convert to pyArrow DataFrame"""
        return pa.Table.from_pandas(
            self.as_pandas().reset_index(drop=True),
            preserve_index=False,
            schema=self.schema.pa_schema,
            safe=type_safe,
        )

    @abstractmethod
    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:  # pragma: no cover
        """Convert to 2-dimensional native python array

        :param columns: columns to extract, defaults to None
        :param type_safe: whether to ensure output conforms with its schema,
          defaults to False
        :return: 2-dimensional native python array

        .. note::

            If ``type_safe`` is False, then the returned values are 'raw' values.
        """
        raise NotImplementedError

    @abstractmethod
    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:  # pragma: no cover
        """Convert to iterable of native python arrays

        :param columns: columns to extract, defaults to None
        :param type_safe: whether to ensure output conforms with its schema,
          defaults to False
        :return: iterable of native python arrays

        .. note::

            If ``type_safe`` is False, then the returned values are 'raw' values.
        """

        raise NotImplementedError

    @abstractmethod
    def _drop_cols(self, cols: List[str]) -> "DataFrame":  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def rename(self, columns: Dict[str, str]) -> "DataFrame":  # pragma: no cover
        """Rename the dataframe using a mapping dict

        :param columns: key: the original column name, value: the new name
        :return: a new dataframe with the new names
        """
        raise NotImplementedError

    @abstractmethod
    def alter_columns(self, columns: Any) -> "DataFrame":  # pragma: no cover
        """Change column types

        :param columns: |SchemaLikeObject|,
          all columns should be contained by the dataframe schema
        :return: a new dataframe with altered columns, the order of the
          original schema will not change
        """
        raise NotImplementedError

    @abstractmethod
    def _select_cols(self, cols: List[Any]) -> "DataFrame":  # pragma: no cover
        raise NotImplementedError

    def drop(self, columns: List[str]) -> "DataFrame":
        """Drop certain columns and return a new dataframe

        :param columns: columns to drop
        :raises FugueDataFrameOperationError: if
          ``columns`` are not strictly contained by this dataframe, or it is the
          entire dataframe columns
        :return: a new dataframe removing the columns
        """
        try:
            schema = self.schema - columns
        except Exception as e:
            raise FugueDataFrameOperationError from e
        if len(schema) == 0:
            raise FugueDataFrameOperationError(
                "can't remove all columns of a dataframe"
            )
        return self._drop_cols(columns)

    def __getitem__(self, columns: List[Any]) -> "DataFrame":
        """Get certain columns of the dataframe as a new dataframe

        :raises FugueDataFrameOperationError: if ``columns`` are not strictly contained
          by this dataframe or it is empty
        :return: a new dataframe with these columns
        """
        try:
            schema = self.schema.extract(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        if len(schema) == 0:
            raise FugueDataFrameOperationError("must select at least one column")
        return self._select_cols(columns)

    @abstractmethod
    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> "LocalBoundedDataFrame":  # pragma: no cover
        """Get first n rows of the dataframe as a new local bounded dataframe

        :param n: number of rows
        :param columns: selected columns, defaults to None (all columns)
        :return: a local bounded dataframe
        """
        raise NotImplementedError

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        """Convert to iterable of native python dicts

        :param columns: columns to extract, defaults to None
        :return: iterable of native python dicts

        .. note::

            The default implementation enforces ``type_safe`` True
        """
        if columns is None:
            columns = self.columns
        idx = range(len(columns))
        for x in self.as_array_iterable(columns, type_safe=True):
            yield {columns[i]: x[i] for i in idx}

    def get_info_str(self) -> str:
        """Get dataframe information (schema, type, metadata) as json string

        :return: json string
        """
        return json.dumps(
            {
                "schema": str(self.schema),
                "type": "{}.{}".format(
                    self.__class__.__module__, self.__class__.__name__
                ),
                "metadata": self.metadata if self.has_metadata else {},
            }
        )

    def __copy__(self) -> "DataFrame":
        return self

    def __deepcopy__(self, memo: Any) -> "DataFrame":
        return self

    def _get_altered_schema(self, subschema: Any) -> Schema:
        sub = Schema(subschema)
        assert_or_throw(
            sub.names in self.schema,
            lambda: FugueDataFrameOperationError(
                f"{sub.names} are not all in {self.schema}"
            ),
        )
        for k, v in sub.items():
            old_type = self.schema[k].type
            new_type = v.type
            if not old_type.equals(new_type):
                assert_or_throw(
                    not pa.types.is_struct(old_type)
                    and not pa.types.is_list(old_type)
                    and not pa.types.is_binary(old_type),
                    lambda: NotImplementedError(f"can't convert from {old_type}"),
                )
                assert_or_throw(
                    not pa.types.is_struct(new_type)
                    and not pa.types.is_list(new_type)
                    and not pa.types.is_binary(new_type),
                    lambda: NotImplementedError(f"can't convert to {new_type}"),
                )
        return Schema([(k, sub.get(k, v)) for k, v in self.schema.items()])


class LocalDataFrame(DataFrame):
    """Base class of all local dataframes. Please read
    :ref:`this <tutorial:tutorials/advanced/schema_dataframes:dataframe>`
    to understand the concept

    :param schema: a `schema-like <triad.collections.schema.Schema>`_ object

    .. note::

        This is an abstract class, and normally you don't construct it by yourself
        unless you are
        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    def native_as_df(self) -> AnyDataFrame:
        return self.as_pandas()

    @property
    def is_local(self) -> bool:
        """Always True because it's a LocalDataFrame"""
        return True

    @property
    def num_partitions(self) -> int:  # pragma: no cover
        """Always 1 because it's a LocalDataFrame"""
        return 1


class LocalBoundedDataFrame(LocalDataFrame):
    """Base class of all local bounded dataframes. Please read
    :ref:`this <tutorial:tutorials/advanced/schema_dataframes:dataframe>`
    to understand the concept

    :param schema: |SchemaLikeObject|

    .. note::

        This is an abstract class, and normally you don't construct it by yourself
        unless you are
        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    @property
    def is_bounded(self) -> bool:
        """Always True because it's a bounded dataframe"""
        return True

    def as_local_bounded(self) -> "LocalBoundedDataFrame":
        """Always True because it's a bounded dataframe"""
        return self


class LocalUnboundedDataFrame(LocalDataFrame):
    """Base class of all local unbounded dataframes. Read
    this <https://fugue-tutorials.readthedocs.io/
    en/latest/tutorials/advanced/schema_dataframes.html#DataFrame>`_
    to understand the concept

    :param schema: |SchemaLikeObject|

    .. note::

        This is an abstract class, and normally you don't construct it by yourself
        unless you are
        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    """

    @property
    def is_bounded(self):
        """Always False because it's an unbounded dataframe"""
        return False

    def as_local(self) -> "LocalDataFrame":
        return self

    def count(self) -> int:
        """
        :raises InvalidOperationError: You can't count an unbounded dataframe
        """
        raise InvalidOperationError("Impossible to count an LocalUnboundedDataFrame")


class YieldedDataFrame(Yielded):
    """Yielded dataframe from :class:`~fugue.workflow.workflow.FugueWorkflow`.
    Users shouldn't create this object directly.

    :param yid: unique id for determinism
    """

    def __init__(self, yid: str):
        super().__init__(yid)
        self._df: Any = None

    @property
    def is_set(self) -> bool:
        return self._df is not None

    def set_value(self, df: DataFrame) -> None:
        """Set the yielded dataframe after compute. Users should not
        call it.

        :param path: file path
        """
        self._df = df

    @property
    def result(self) -> DataFrame:
        """The yielded dataframe, it will be set after the parent
        workflow is computed
        """
        assert_or_throw(self.is_set, "value is not set")
        return self._df


class DataFrameDisplay(DatasetDisplay):
    """:class:`~.DataFrame` plain display class"""

    @property
    def df(self) -> DataFrame:
        """The target :class:`~.DataFrame`"""
        return self._ds  # type: ignore

    def show(
        self, n: int = 10, with_count: bool = False, title: Optional[str] = None
    ) -> None:
        best_width = 100
        head_rows = self.df.head(n).as_array(type_safe=True)
        if len(head_rows) < n:
            count = len(head_rows)
        else:
            count = self.df.count() if with_count else -1
        with DatasetDisplay._SHOW_LOCK:
            if title is not None and title != "":
                print(title)
            print(type(self.df).__name__)
            tb = PrettyTable(self.df.schema, head_rows, best_width)
            print("\n".join(tb.to_string()))
            if count >= 0:
                print(f"Total count: {count}")
                print("")
            if self.df.has_metadata:
                print("Metadata:")
                try:
                    # try pretty print, but if not convertible to json, print original
                    print(self.df.metadata.to_json(indent=True))
                except Exception:  # pragma: no cover
                    print(self.df.metadata)
                print("")


def as_fugue_df(df: AnyDataFrame, **kwargs: Any) -> DataFrame:
    """Wrap the object as a Fugue DataFrame.

    :param df: the object to wrap
    """
    ds = as_fugue_dataset(df, **kwargs)
    if isinstance(ds, DataFrame):
        return ds
    raise TypeError(f"{type(df)} {kwargs} is not recognized as a Fugue DataFrame: {ds}")


@get_dataset_display.candidate(lambda ds: isinstance(ds, DataFrame), priority=0.1)
def _get_dataframe_display(ds: DataFrame):
    return DataFrameDisplay(ds)


@as_local.candidate(lambda df: isinstance(df, DataFrame))
def _df_to_local(df: DataFrame) -> LocalDataFrame:
    return df.as_local()


@as_local_bounded.candidate(lambda df: isinstance(df, DataFrame))
def _df_to_local_bounded(df: DataFrame) -> LocalBoundedDataFrame:
    return df.as_local_bounded()


def _get_schema_change(
    orig_schema: Optional[Schema], schema: Any
) -> Tuple[Schema, List[int]]:
    if orig_schema is None:
        schema = _input_schema(schema).assert_not_empty()
        return schema, []
    elif schema is None:
        return orig_schema.assert_not_empty(), []
    if isinstance(schema, (str, Schema)) and orig_schema == schema:
        return orig_schema.assert_not_empty(), []
    if schema in orig_schema:
        # keys list or schema like object that is a subset of orig
        schema = orig_schema.extract(schema).assert_not_empty()
        pos = [orig_schema.index_of_key(x) for x in schema.names]
        if pos == list(range(len(orig_schema))):
            pos = []
        return schema, pos
    # otherwise it has to be a schema like object that must be a subset
    # of orig, and that has mismatched types
    schema = _input_schema(schema).assert_not_empty()
    pos = [orig_schema.index_of_key(x) for x in schema.names]
    if pos == list(range(len(orig_schema))):
        pos = []
    return schema, pos


def _input_schema(schema: Any) -> Schema:
    return schema if isinstance(schema, Schema) else Schema(schema)



import copy
from typing import Any, Callable, Dict, List, Optional, no_type_check

from triad import ParamDict
from triad.collections import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import get_caller_global_local_vars, to_function, to_instance
from triad.utils.hash import to_uuid

from fugue._utils.interfaceless import parse_output_schema_from_comment
from fugue._utils.registry import fugue_plugin
from fugue.dataframe import DataFrame
from fugue.dataframe.function_wrapper import DataFrameFunctionWrapper
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions.creator.creator import Creator

from .._utils import load_namespace_extensions

_CREATOR_REGISTRY = ParamDict()


@fugue_plugin
def parse_creator(obj: Any) -> Any:
    """Parse an object to another object that can be converted to a Fugue
    :class:`~fugue.extensions.creator.creator.Creator`.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import Creator, FugueWorkflow
            from fugue.plugins import parse_creator
            from triad import to_uuid

            class My(Creator):
                def __init__(self, x):
                    self.x = x

                def create(self) :
                    raise NotImplementedError

                def __uuid__(self) -> str:
                    return to_uuid(super().__uuid__(), self.x)

            @parse_creator.candidate(
                lambda x: isinstance(x, str) and x.startswith("-*"))
            def _parse(obj):
                return My(obj)

            dag = FugueWorkflow()
            dag.create("-*abc").show()
            # == dag.create(My("-*abc")).show()

            dag.run()
    """
    if isinstance(obj, str) and obj in _CREATOR_REGISTRY:
        return _CREATOR_REGISTRY[obj]
    return obj


def register_creator(alias: str, obj: Any, on_dup: int = ParamDict.OVERWRITE) -> None:
    """Register creator with an alias. This is a simplified version of
    :func:`~.parse_creator`

    :param alias: alias of the creator
    :param obj: the object that can be converted to
        :class:`~fugue.extensions.creator.creator.Creator`
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
        , defaults to ``ParamDict.OVERWRITE``

    .. tip::

        Registering an extension with an alias is particularly useful for projects
        such as libraries. This is because by using alias, users don't have to
        import the specific extension, or provide the full path of the extension.
        It can make user's code less dependent and easy to understand.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please read
        :doc:`Creator Tutorial <tutorial:tutorials/extensions/creator>`

    .. admonition:: Examples

        Here is an example how you setup your project so your users can
        benefit from this feature. Assume your project name is ``pn``

        The creator implementation in file ``pn/pn/creators.py``

        .. code-block:: python

            import pandas import pd

            def my_creator() -> pd.DataFrame:
                return pd.DataFrame()

        Then in ``pn/pn/__init__.py``

        .. code-block:: python

            from .creators import my_creator
            from fugue import register_creator

            def register_extensions():
                register_creator("mc", my_creator)
                # ... register more extensions

            register_extensions()

        In users code:

        .. code-block:: python

            import pn  # register_extensions will be called
            from fugue import FugueWorkflow

            dag = FugueWorkflow()
            dag.create("mc").show()  # use my_creator by alias
            dag.run()
    """
    _CREATOR_REGISTRY.update({alias: obj}, on_dup=on_dup)


def creator(schema: Any = None) -> Callable[[Any], "_FuncAsCreator"]:
    """Decorator for creators

    Please read
    :doc:`Creator Tutorial <tutorial:tutorials/extensions/creator>`
    """

    def deco(func: Callable) -> "_FuncAsCreator":
        return _FuncAsCreator.from_func(func, schema)

    return deco


def _to_creator(
    obj: Any,
    schema: Any = None,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
) -> Creator:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    load_namespace_extensions(obj)
    obj = parse_creator(obj)
    exp: Optional[Exception] = None
    try:
        return copy.copy(
            to_instance(obj, Creator, global_vars=global_vars, local_vars=local_vars)
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, Creator):
            return copy.copy(f)
        # this is for functions without decorator
        return _FuncAsCreator.from_func(f, schema)
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid creator", exp)


class _FuncAsCreator(Creator):
    @no_type_check
    def create(self) -> DataFrame:
        args: List[Any] = []
        kwargs: Dict[str, Any] = {}
        if self._engine_param is not None:
            args.append(self._engine_param.to_input(self.execution_engine))
        kwargs.update(self.params)
        return self._wrapper.run(  # type: ignore
            args=args,
            kwargs=kwargs,
            output_schema=self.output_schema if self._need_output_schema else None,
            ctx=self.execution_engine,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(
            self._wrapper,
            self._engine_param,
            self._need_output_schema,
            str(self._output_schema),
        )

    @no_type_check
    @staticmethod
    def from_func(func: Callable, schema: Any) -> "_FuncAsCreator":
        # pylint: disable=W0201
        if schema is None:
            schema = parse_output_schema_from_comment(func)
        tr = _FuncAsCreator()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^e?x*z?$", "^[dlspq]$"
        )
        tr._engine_param = (
            tr._wrapper._params.get_value_by_index(0)
            if tr._wrapper.input_code.startswith("e")
            else None
        )
        tr._need_output_schema = tr._wrapper.need_output_schema
        tr._output_schema = Schema(schema)
        if len(tr._output_schema) == 0:
            assert_or_throw(
                tr._need_output_schema is None or not tr._need_output_schema,
                FugueInterfacelessError(
                    f"schema must be provided for return type {tr._wrapper._rt}"
                ),
            )
        else:
            assert_or_throw(
                tr._need_output_schema is None or tr._need_output_schema,
                FugueInterfacelessError(
                    f"schema must not be provided for return type {tr._wrapper._rt}"
                ),
            )
        return tr



from typing import Any, Iterable, List

from triad import Schema


class PrettyTable:
    def __init__(
        self,  # noqa: C901
        schema: Schema,
        data: List[Any],
        best_width: int,
        truncate_width: int = 500,
    ):
        raw: List[List[str]] = []
        self.titles = str(schema).split(",")
        col_width_min = [len(t) for t in self.titles]
        col_width_max = list(col_width_min)
        self.col_width = list(col_width_min)
        # Convert all cells to string with truncation
        for row in data:
            raw_row: List[str] = []
            for i in range(len(schema)):
                d = self._cell_to_raw_str(row[i], truncate_width)
                col_width_max[i] = max(col_width_max[i], len(d))
                raw_row.append(d)
            raw.append(raw_row)
        # Adjust col width based on best_width
        # It find the remaining width after fill all cols with min widths,
        # and redistribute the remaining based on the diff between max and min widths
        dt = sorted(
            filter(  # noqa: C407
                lambda x: x[0] > 0,
                [(w - col_width_min[i], i) for i, w in enumerate(col_width_max)],
            )
        )
        if len(dt) > 0:
            remaining = max(0, best_width - sum(col_width_min) - len(col_width_min)) + 1
            total = sum(x[0] for x in dt)
            for diff, index in dt:
                if remaining <= 0:  # pragma: no cover
                    break
                given = remaining * diff // total
                remaining -= given
                self.col_width[index] += given
        # construct data -> List[List[List[str]]], make sure on the same row, each cell
        # has the same length of strings
        self.data = [
            [self._wrap(row[i], self.col_width[i]) for i in range(len(schema))]
            for row in raw
        ]
        blank = ["".ljust(self.col_width[i]) for i in range(len(schema))]
        for row in self.data:
            max_h = max(len(c) for c in row)
            for i in range(len(schema)):
                row[i] += [blank[i]] * (max_h - len(row[i]))

    def to_string(self) -> Iterable[str]:
        yield "|".join(
            self.titles[i].ljust(self.col_width[i]) for i in range(len(self.titles))
        )
        yield "+".join(
            "".ljust(self.col_width[i], "-") for i in range(len(self.titles))
        )
        for row in self.data:
            for tp in zip(*row):
                yield "|".join(tp)

    def _cell_to_raw_str(self, obj: Any, truncate_width: int) -> str:
        raw = "NULL" if obj is None else str(obj)
        if len(raw) > truncate_width:
            raw = raw[: max(0, truncate_width - 3)] + "..."
        return raw

    def _wrap(self, s: str, width: int) -> List[str]:
        res: List[str] = []
        start = 0
        while start < len(s):
            end = min(len(s), start + width)
            sub = s[start:end]
            if end < len(s):
                res.append(sub)
            else:
                res.append(sub.ljust(width, " "))
                break
            start += width
        return res



import copy
from typing import Any, Callable, Dict, List, Optional, no_type_check

from triad import ParamDict, to_uuid
from triad.collections import Schema
from triad.utils.assertion import assert_or_throw
from triad.utils.convert import get_caller_global_local_vars, to_function, to_instance

from fugue._utils.interfaceless import parse_output_schema_from_comment
from fugue._utils.registry import fugue_plugin
from fugue.dataframe import DataFrame, DataFrames
from fugue.dataframe.function_wrapper import DataFrameFunctionWrapper
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions.processor.processor import Processor

from .._utils import (
    load_namespace_extensions,
    parse_validation_rules_from_comment,
    to_validation_rules,
)

_PROCESSOR_REGISTRY = ParamDict()


@fugue_plugin
def parse_processor(obj: Any) -> Any:
    """Parse an object to another object that can be converted to a Fugue
    :class:`~fugue.extensions.processor.processor.Processor`.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import Processor, parse_processor, FugueWorkflow
            from triad import to_uuid

            class My(Processor):
                def __init__(self, x):
                    self.x = x

                def process(self, dfs):
                    raise NotImplementedError

                def __uuid__(self) -> str:
                    return to_uuid(super().__uuid__(), self.x)

            @parse_processor.candidate(
                lambda x: isinstance(x, str) and x.startswith("-*"))
            def _parse(obj):
                return My(obj)

            dag = FugueWorkflow()
            dag.df([[0]], "a:int").process("-*abc")
            # ==  dag.df([[0]], "a:int").process(My("-*abc"))

            dag.run()
    """
    if isinstance(obj, str) and obj in _PROCESSOR_REGISTRY:
        return _PROCESSOR_REGISTRY[obj]
    return obj


def register_processor(alias: str, obj: Any, on_dup: int = ParamDict.OVERWRITE) -> None:
    """Register processor with an alias.

    :param alias: alias of the processor
    :param obj: the object that can be converted to
        :class:`~fugue.extensions.processor.processor.Processor`
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
        , defaults to ``ParamDict.OVERWRITE``

    .. tip::

        Registering an extension with an alias is particularly useful for projects
        such as libraries. This is because by using alias, users don't have to
        import the specific extension, or provide the full path of the extension.
        It can make user's code less dependent and easy to understand.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please read
        :doc:`Processor Tutorial <tutorial:tutorials/extensions/processor>`

    .. admonition:: Examples

        Here is an example how you setup your project so your users can
        benefit from this feature. Assume your project name is ``pn``

        The processor implementation in file ``pn/pn/processors.py``

        .. code-block:: python

            from fugue import DataFrame

            def my_processor(df:DataFrame) -> DataFrame:
                return df

        Then in ``pn/pn/__init__.py``

        .. code-block:: python

            from .processors import my_processor
            from fugue import register_processor

            def register_extensions():
                register_processor("mp", my_processor)
                # ... register more extensions

            register_extensions()

        In users code:

        .. code-block:: python

            import pn  # register_extensions will be called
            from fugue import FugueWorkflow

            dag = FugueWorkflow()
            # use my_processor by alias
            dag.df([[0]],"a:int").process("mp").show()
            dag.run()
    """
    _PROCESSOR_REGISTRY.update({alias: obj}, on_dup=on_dup)


def processor(
    schema: Any = None, **validation_rules: Any
) -> Callable[[Any], "_FuncAsProcessor"]:
    """Decorator for processors

    Please read
    :doc:`Processor Tutorial <tutorial:tutorials/extensions/processor>`
    """
    # TODO: validation of schema if without * should be done at compile time
    def deco(func: Callable) -> "_FuncAsProcessor":
        return _FuncAsProcessor.from_func(
            func, schema, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def _to_processor(
    obj: Any,
    schema: Any = None,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    validation_rules: Optional[Dict[str, Any]] = None,
) -> Processor:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    load_namespace_extensions(obj)
    obj = parse_processor(obj)
    exp: Optional[Exception] = None
    if validation_rules is None:
        validation_rules = {}
    try:
        return copy.copy(
            to_instance(obj, Processor, global_vars=global_vars, local_vars=local_vars)
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, Processor):
            return copy.copy(f)
        # this is for functions without decorator
        return _FuncAsProcessor.from_func(f, schema, validation_rules=validation_rules)
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid processor", exp)


class _FuncAsProcessor(Processor):
    @property
    def validation_rules(self) -> Dict[str, Any]:
        return self._validation_rules  # type: ignore

    @no_type_check
    def process(self, dfs: DataFrames) -> DataFrame:
        args: List[Any] = []
        kwargs: Dict[str, Any] = {}
        if self._engine_param is not None:
            args.append(self._engine_param.to_input(self.execution_engine))
        if self._use_dfs:
            args.append(dfs)
        else:
            if not dfs.has_key:
                args += dfs.values()
            else:
                kwargs.update(dfs)
        kwargs.update(self.params)
        return self._wrapper.run(
            args=args,
            kwargs=kwargs,
            output_schema=self.output_schema if self._need_output_schema else None,
            ctx=self.execution_engine,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(
            self._wrapper,
            self._engine_param,
            self._use_dfs,
            self._need_output_schema,
            str(self._output_schema),
        )

    @no_type_check
    @staticmethod
    def from_func(
        func: Callable, schema: Any, validation_rules: Dict[str, Any]
    ) -> "_FuncAsProcessor":
        if schema is None:
            schema = parse_output_schema_from_comment(func)
        validation_rules.update(parse_validation_rules_from_comment(func))
        tr = _FuncAsProcessor()
        tr._wrapper = DataFrameFunctionWrapper(
            func, "^e?(c|[dlspq]+)x*z?$", "^[dlspq]$"
        )  # type: ignore
        tr._engine_param = (
            tr._wrapper._params.get_value_by_index(0)
            if tr._wrapper.input_code.startswith("e")
            else None
        )
        tr._use_dfs = "c" in tr._wrapper.input_code
        tr._need_output_schema = tr._wrapper.need_output_schema
        tr._validation_rules = validation_rules
        tr._output_schema = Schema(schema)
        if len(tr._output_schema) == 0:
            assert_or_throw(
                tr._need_output_schema is None or not tr._need_output_schema,
                FugueInterfacelessError(
                    f"schema must be provided for return type {tr._wrapper._rt}"
                ),
            )
        else:
            assert_or_throw(
                tr._need_output_schema is None or tr._need_output_schema,
                FugueInterfacelessError(
                    f"schema must not be provided for return type {tr._wrapper._rt}"
                ),
            )
        return tr



from datetime import date, datetime
from typing import Any, Dict, Iterable, Optional, Tuple

import numpy as np
import pandas as pd
import pyarrow as pa
from triad import Schema
from triad.utils.pyarrow import TRIAD_DEFAULT_TIMESTAMP
from triad.utils.schema import quote_name

_DUCK_TYPES_TO_PA: Dict[str, pa.DataType] = {
    "BIGINT": pa.int64(),
    "BOOLEAN": pa.bool_(),
    "BLOB": pa.binary(),
    "DATE": pa.date32(),
    "DOUBLE": pa.float64(),
    "INTEGER": pa.int32(),
    "FLOAT": pa.float32(),
    "SMALLINT": pa.int16(),
    "TIMESTAMP": TRIAD_DEFAULT_TIMESTAMP,
    "TINYINT": pa.int8(),
    "UBIGINT": pa.uint64(),
    "UINTEGER": pa.uint32(),
    "USMALLINT": pa.uint16(),
    "UTINYINT": pa.uint8(),
    "VARCHAR": pa.string(),
    "TIME": pa.time32("ms"),
}

_PA_TYPES_TO_DUCK: Dict[pa.DataType, str] = {
    v: k
    for k, v in list(_DUCK_TYPES_TO_PA.items())
    + [("VARCHAR", pa.large_string()), ("BLOB", pa.large_binary())]
}


def encode_column_name(name: str) -> str:
    return quote_name(name, quote='"')


def encode_column_names(names: Iterable[str]) -> Iterable[str]:
    for name in names:
        yield encode_column_name(name)


def encode_schema_names(schema: Schema) -> Iterable[str]:
    return encode_column_names(schema.names)


def encode_value_to_expr(value: Any) -> str:  # noqa: C901
    if isinstance(value, list):
        return "[" + ", ".join(encode_value_to_expr(x) for x in value) + "]"
    if isinstance(value, dict):
        return (
            "{"
            + ", ".join(
                encode_value_to_expr(k) + ": " + encode_value_to_expr(v)
                for k, v in value.items()
            )
            + "}"
        )
    if pd.isna(value):
        return "NULL"
    if isinstance(value, np.generic):
        value = value.item()
    if isinstance(value, str):
        return "E" + repr(value)
    if isinstance(value, bytes):
        return repr(value)[1:] + "::BLOB"
    if isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, datetime):
        return "TIMESTAMP '" + value.strftime("%Y-%m-%d %H:%M:%S") + "'"
    if isinstance(value, date):
        return "DATE '" + value.strftime("%Y-%m-%d") + "'"

    raise NotImplementedError(value)


def to_duck_type(tp: pa.DataType) -> str:
    try:
        if pa.types.is_struct(tp):
            inner = ",".join(f.name + " " + to_duck_type(f.type) for f in tp)
            return f"STRUCT({inner})"
        if pa.types.is_map(tp):
            k = to_duck_type(tp.key_type)
            v = to_duck_type(tp.item_type)
            return f"MAP({k},{v})"
        if pa.types.is_list(tp):
            inner = to_duck_type(tp.value_type)
            return f"{inner}[]"
        if pa.types.is_decimal(tp):
            return f"DECIMAL({tp.precision}, {tp.scale})"
        return _PA_TYPES_TO_DUCK[tp]
    except Exception:  # pragma: no cover
        raise ValueError(f"can't convert {tp} to DuckDB data type")


def to_pa_type(duck_type_raw: Any) -> pa.DataType:
    try:
        duck_type = str(duck_type_raw)  # for duckdb >= 0.8.0
        if duck_type.endswith("[]"):
            return pa.list_(to_pa_type(duck_type[:-2]))
        p = duck_type.find("(")
        if p > 0:
            tp = duck_type[:p]
            if tp == "STRUCT":
                fields = [
                    pa.field(k, to_pa_type(v.strip()))
                    for k, v in _split_comma(duck_type[p + 1 : -1])
                ]
                return pa.struct(fields)
            if tp == "MAP":
                fields = [
                    to_pa_type(t.strip())
                    for t, _ in _split_comma(duck_type[p + 1 : -1], split_char=None)
                ]
                return pa.map_(fields[0], fields[1])
            if tp != "DECIMAL":
                raise Exception
            pair = duck_type[p + 1 : -1].split(",", 1)
            return pa.decimal128(int(pair[0]), int(pair[1]))
        if duck_type == "HUGEINT":
            return pa.int64()
        if duck_type in _DUCK_TYPES_TO_PA:
            return _DUCK_TYPES_TO_PA[duck_type]
        raise Exception
    except Exception:
        raise ValueError(f"{duck_type} is not supported")


def _split_comma(
    expr: str,
    split_char: Optional[str] = " ",
    left_char: str = "(",
    right_char: str = ")",
) -> Iterable[Tuple[str, str]]:
    lv = 0
    start = 0
    for i in range(len(expr)):
        if expr[i] == left_char:
            lv += 1
        elif expr[i] == right_char:
            lv -= 1
        elif lv == 0 and expr[i] == ",":
            if split_char is None:
                yield expr[start:i].strip(), ""
            else:
                x = expr[start:i].strip().split(split_char, 1)
                yield x[0], x[1]
            start = i + 1
    if split_char is None:
        yield expr[start : len(expr)].strip(), ""
    else:
        x = expr[start : len(expr)].strip().split(split_char, 1)
        yield x[0], x[1]



from typing import Any, Dict, Iterable, List, Optional, Union

import pyarrow as pa
from triad import Schema, assert_or_throw, to_uuid
from triad.utils.pyarrow import _type_to_expression, to_pa_datatype


class ColumnExpr:
    """Fugue column expression class. It is inspired from
    :class:`spark:pyspark.sql.Column` and it is working in progress.

    .. admonition:: New Since
            :class: hint

            **0.6.0**

    .. caution::

        This is a base class of different column classes, and users are not supposed
        to construct this class directly. Use :func:`~.col` and :func:`~.lit` instead.
    """

    def __init__(self):
        self._as_name = ""
        self._as_type: Optional[pa.DataType] = None

    @property
    def name(self) -> str:
        """The original name of this column, default empty

        :return: the name

        .. admonition:: Examples

            .. code-block:: python

                assert "a" == col("a").name
                assert "b" == col("a").alias("b").name
                assert "" == lit(1).name
                assert "" == (col("a") * 2).name
        """
        return ""

    @property
    def as_name(self) -> str:
        """The name assigned by :meth:`~.alias`

        :return: the alias

        .. admonition:: Examples

            .. code-block:: python

                assert "" == col("a").as_name
                assert "b" == col("a").alias("b").as_name
                assert "x" == (col("a") * 2).alias("x").as_name
        """
        return self._as_name

    @property
    def as_type(self) -> Optional[pa.DataType]:
        """The type assigned by :meth:`~.cast`

        :return: the pyarrow datatype if :meth:`~.cast` was called
          otherwise None

        .. admonition:: Examples

            .. code-block:: python

                import pyarrow as pa

                assert col("a").as_type is None
                assert pa.int64() == col("a").cast(int).as_type
                assert pa.string() == (col("a") * 2).cast(str).as_type
        """
        return self._as_type

    @property
    def output_name(self) -> str:
        """The name assigned by :meth:`~.alias`, but if empty then
        return the original column name

        :return: the alias or the original column name

        .. admonition:: Examples

            .. code-block:: python

                assert "a" == col("a").output_name
                assert "b" == col("a").alias("b").output_name
                assert "x" == (col("a") * 2).alias("x").output_name
        """
        return self.as_name if self.as_name != "" else self.name

    def alias(self, as_name: str) -> "ColumnExpr":  # pragma: no cover
        """Assign or remove alias of a column. To remove, set ``as_name`` to empty

        :return: a new column with the alias value

        .. admonition:: Examples

            .. code-block:: python

                assert "b" == col("a").alias("b").as_name
                assert "x" == (col("a") * 2).alias("x").as_name
                assert "" == col("a").alias("b").alias("").as_name
        """
        raise NotImplementedError

    def infer_alias(self) -> "ColumnExpr":
        """Infer alias of a column. If the column's :meth:`~.output_name` is not empty
        then it returns itself without change. Otherwise it tries to infer alias from
        the underlying columns.

        :return: a column instance with inferred alias

        .. caution::

            Users should not use it directly.

        .. admonition:: Examples

            .. code-block:: python

                import fugue.column.functions as f

                assert "a" == col("a").infer_alias().output_name
                assert "" == (col("a") * 2).infer_alias().output_name
                assert "a" == col("a").is_null().infer_alias().output_name
                assert "a" == f.max(col("a").is_null()).infer_alias().output_name
        """
        return self

    def cast(self, data_type: Any) -> "ColumnExpr":  # pragma: no cover
        """Cast the column to a new data type

        :param data_type: It can be string expressions, python primitive types,
          python `datetime.datetime` and pyarrow types.
          For details read |FugueDataTypes|
        :return: a new column instance with the assigned data type

        .. caution::

            Currently, casting to struct or list type has undefined behavior.

        .. admonition:: Examples

            .. code-block:: python

                import pyarrow as pa

                assert pa.int64() == col("a").cast(int).as_type
                assert pa.string() == col("a").cast(str).as_type
                assert pa.float64() == col("a").cast(float).as_type
                assert pa._bool() == col("a").cast(bool).as_type

                # string follows the type expression of Triad Schema
                assert pa.int32() == col("a").cast("int").as_type
                assert pa.int32() == col("a").cast("int32").as_type

                assert pa.int32() == col("a").cast(pa.int32()).as_type
        """
        raise NotImplementedError

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        """Infer data type of this column given the input schema

        :param schema: the schema instance to infer from
        :return: a pyarrow datatype or None if failed to infer

        .. caution::

            Users should not use it directly.

        .. admonition:: Examples

            .. code-block:: python

                import pyarrow as pa
                from triad import Schema
                import fugue.column.functions as f

                schema = Schema("a:int,b:str")

                assert pa.int32() == col("a").infer_schema(schema)
                assert pa.int32() == (-col("a")).infer_schema(schema)
                # due to overflow risk, can't infer certain operations
                assert (col("a")+1).infer_schema(schema) is None
                assert (col("a")+col("a")).infer_schema(schema) is None
                assert pa.int32() == f.max(col("a")).infer_schema(schema)
                assert pa.int32() == f.min(col("a")).infer_schema(schema)
                assert f.sum(col("a")).infer_schema(schema) is None
        """
        return self.as_type  # pragma: no cover

    def __str__(self) -> str:
        """String expression of the column, this is only used for debug purpose.
        It is not SQL expression.

        :return: the string expression
        """
        res = self.body_str
        if self.as_type is not None:
            res = f"CAST({res} AS {_type_to_expression(self.as_type)})"
        if self.as_name != "":
            res = res + " AS " + self.as_name
        return res

    @property
    def body_str(self) -> str:  # pragma: no cover
        """The string expression of this column without cast type and alias.
        This is only used for debug purpose. It is not SQL expression.

        :return: the string expression
        """
        raise NotImplementedError

    def is_null(self) -> "ColumnExpr":
        """Same as SQL ``<col> IS NULL``.

        :return: a new column with the boolean values
        """
        # TODO: should enable infer_schema for this?
        return _UnaryOpExpr("IS_NULL", self)

    def not_null(self) -> "ColumnExpr":
        """Same as SQL ``<col> IS NOT NULL``.

        :return: a new column with the boolean values
        """
        # TODO: should enable infer_schema for this?
        return _UnaryOpExpr("NOT_NULL", self)

    def __neg__(self) -> "ColumnExpr":
        """The negative value of the current column

        :return: a new column with the negative value
        """
        return _InvertOpExpr("-", self)

    def __pos__(self) -> "ColumnExpr":
        """The original value of the current column

        :return: the column itself
        """
        return self

    def __invert__(self) -> "ColumnExpr":
        """Same as SQL ``NOT <col>``

        :return: a new column with the boolean values
        """
        return _NotOpExpr("~", self)

    def __add__(self, other: Any) -> "ColumnExpr":
        """Add with another column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("+", self, other)

    def __radd__(self, other: Any) -> "ColumnExpr":
        """Add with another column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("+", other, self)

    def __sub__(self, other: Any) -> "ColumnExpr":
        """Subtract another column from this column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("-", self, other)

    def __rsub__(self, other: Any) -> "ColumnExpr":
        """Subtract this column from the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("-", other, self)

    def __mul__(self, other: Any) -> "ColumnExpr":
        """Multiply with another column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("*", self, other)

    def __rmul__(self, other: Any) -> "ColumnExpr":
        """Multiply with another column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("*", other, self)

    def __truediv__(self, other: Any) -> "ColumnExpr":
        """Divide this column by the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("/", self, other)

    def __rtruediv__(self, other: Any) -> "ColumnExpr":
        """Divide the other column by this column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BinaryOpExpr("/", other, self)

    def __and__(self, other: Any) -> "ColumnExpr":
        """``AND`` value of the two columns

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BoolBinaryOpExpr("&", self, other)

    def __rand__(self, other: Any) -> "ColumnExpr":
        """``AND`` value of the two columns

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BoolBinaryOpExpr("&", other, self)

    def __or__(self, other: Any) -> "ColumnExpr":
        """``OR`` value of the two columns

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BoolBinaryOpExpr("|", self, other)

    def __ror__(self, other: Any) -> "ColumnExpr":
        """``OR`` value of the two columns

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the result
        """
        return _BoolBinaryOpExpr("|", other, self)

    def __lt__(self, other: Any) -> "ColumnExpr":
        """Whether this column is less than the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr("<", self, other)

    def __gt__(self, other: Any) -> "ColumnExpr":
        """Whether this column is greater than the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr(">", self, other)

    def __le__(self, other: Any) -> "ColumnExpr":
        """Whether this column is less or equal to the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr("<=", self, other)

    def __ge__(self, other: Any) -> "ColumnExpr":
        """Whether this column is greater or equal to the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr(">=", self, other)

    def __eq__(self, other: Any) -> "ColumnExpr":  # type: ignore
        """Whether this column equals the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr("==", self, other)

    def __ne__(self, other: Any) -> "ColumnExpr":  # type: ignore
        """Whether this column does not equal the other column

        :param other: the other column, if it is not a
          :class:`~.ColumnExpr`, then the value will be converted to
          a literal (`lit(other)`)
        :return: a new column with the boolean result
        """
        return _BoolBinaryOpExpr("!=", self, other)

    def __uuid__(self) -> str:
        """The unique id of this instance

        :return: the unique id
        """
        return to_uuid(
            str(type(self)),
            self.as_name,
            self.as_type,
            *self._uuid_keys(),
        )

    def _uuid_keys(self) -> List[Any]:  # pragma: no cover
        raise NotImplementedError


def lit(obj: Any, alias: str = "") -> ColumnExpr:
    """Convert the ``obj`` to a literal column. Currently ``obj`` must be
    ``int``, ``bool``, ``float`` or ``str``, otherwise an exception will
    be raised

    :param obj: an arbitrary value
    :param alias: the alias of this literal column, defaults to "" (no alias)
    :return: a literal column expression

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column import lit

            lit("abc")
            lit(100).alias("x")
            lit(100, "x")
    """
    return (
        _LiteralColumnExpr(obj) if alias == "" else _LiteralColumnExpr(obj).alias(alias)
    )


def null() -> ColumnExpr:
    """Equivalent to ``lit(None)``, the ``NULL`` value

    :return: ``lit(None)``

    .. admonition:: New Since
        :class: hint

        **0.6.0**
    """
    return lit(None)


def col(obj: Union[str, ColumnExpr], alias: str = "") -> ColumnExpr:
    """Convert the ``obj`` to a :class:`~.ColumnExpr` object

    :param obj: a string representing a column name or a :class:`~.ColumnExpr` object
    :param alias: the alias of this column, defaults to "" (no alias)
    :return: a literal column expression

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. admonition:: Examples

        .. code-block:: python

            import fugue.column import col
            import fugue.column.functions as f

            col("a")
            col("a").alias("x")
            col("a", "x")

            # unary operations
            -col("a")  # negative value of a
            ~col("a")  # NOT a
            col("a").is_null()  # a IS NULL
            col("a").not_null()  # a IS NOT NULL

            # binary operations
            col("a") + 1  # col("a") + lit(1)
            1 - col("a")  # lit(1) - col("a")
            col("a") * col("b")
            col("a") / col("b")

            # binary boolean expressions
            col("a") == 1  # col("a") == lit(1)
            2 != col("a")  # col("a") != lit(2)
            col("a") < 5
            col("a") > 5
            col("a") <= 5
            col("a") >= 5
            (col("a") < col("b")) & (col("b") > 1) | col("c").is_null()

            # with functions
            f.max(col("a"))
            f.max(col("a")+col("b"))
            f.max(col("a")) + f.min(col("b"))
            f.count_distinct(col("a")).alias("dcount")

    """
    if isinstance(obj, ColumnExpr):
        return obj if alias == "" else obj.alias(alias)
    if isinstance(obj, str):
        return (
            _NamedColumnExpr(obj) if alias == "" else _NamedColumnExpr(obj).alias(alias)
        )
    raise NotImplementedError(obj)


def all_cols() -> ColumnExpr:
    """The ``*`` expression in SQL"""
    return _WildcardExpr()


def function(name: str, *args: Any, arg_distinct: bool = False, **kwargs) -> ColumnExpr:
    """Construct a function expression

    :param name: the name of the function
    :param arg_distinct: whether to add ``DISTINCT`` before all arguments,
      defaults to False
    :return: the function expression

    .. caution::

        Users should not use this directly

    """
    return _FuncExpr(name, *args, arg_distinct=arg_distinct, **kwargs)


def _get_column_mentions(column: ColumnExpr) -> Iterable[str]:
    if isinstance(column, _NamedColumnExpr):
        yield column.name
    elif isinstance(column, _FuncExpr):
        for a in column.args:
            yield from _get_column_mentions(a)
        for a in column.kwargs.values():
            yield from _get_column_mentions(a)


def _to_col(obj: Any) -> ColumnExpr:
    if isinstance(obj, ColumnExpr):
        return obj
    return lit(obj)


class _NamedColumnExpr(ColumnExpr):
    def __init__(self, name: Any):
        self._name = name
        super().__init__()

    @property
    def body_str(self) -> str:
        return self.name

    @property
    def name(self) -> str:
        return self._name

    @property
    def output_name(self) -> str:
        return super().output_name

    def alias(self, as_name: str) -> ColumnExpr:
        other = _NamedColumnExpr(self.name)
        other._as_name = as_name
        other._as_type = self.as_type
        return other

    def cast(self, data_type: Any) -> "ColumnExpr":
        other = _NamedColumnExpr(self.name)
        other._as_name = self.as_name
        other._as_type = None if data_type is None else to_pa_datatype(data_type)
        return other

    def infer_alias(self) -> ColumnExpr:
        if self.as_name == "" and self.as_type is not None:
            return self.alias(self.output_name)
        return self

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        if self.name not in schema:
            return self.as_type
        return self.as_type or schema[self.name].type

    def _uuid_keys(self) -> List[Any]:
        return [self.name]


class _WildcardExpr(ColumnExpr):
    @property
    def body_str(self) -> str:
        return "*"

    @property
    def name(self) -> str:  # pragma: no cover
        raise NotImplementedError("wildcard column doesn't have a name")

    @property
    def output_name(self) -> str:  # pragma: no cover
        raise NotImplementedError("wildcard column doesn't have an output_name")

    def __uuid__(self) -> str:
        """The unique id of this instance

        :return: the unique id
        """
        return to_uuid(str(type(self)))


class _LiteralColumnExpr(ColumnExpr):
    _VALID_TYPES = (int, bool, float, str)

    def __init__(self, value: Any):
        assert_or_throw(
            value is None or isinstance(value, _LiteralColumnExpr._VALID_TYPES),
            lambda: NotImplementedError(f"{value}, type: {type(value)}"),
        )
        self._value = value
        super().__init__()

    @property
    def body_str(self) -> str:
        if self.value is None:
            return "NULL"
        elif isinstance(self.value, str):
            body = self.value.translate(
                str.maketrans(
                    {  # type: ignore
                        "\\": r"\\",
                        "'": r"\'",
                    }
                )
            )
            return f"'{body}'"
        elif isinstance(self.value, bool):
            return "TRUE" if self.value else "FALSE"
        else:
            return str(self.value)

    @property
    def value(self) -> Any:
        return self._value

    def is_null(self) -> ColumnExpr:
        return _LiteralColumnExpr(self.value is None)

    def not_null(self) -> ColumnExpr:
        return _LiteralColumnExpr(self.value is not None)

    def alias(self, as_name: str) -> ColumnExpr:
        other = _LiteralColumnExpr(self.value)
        other._as_name = as_name
        other._as_type = self.as_type
        return other

    def cast(self, data_type: Any) -> ColumnExpr:
        other = _LiteralColumnExpr(self.value)
        other._as_name = self.as_name
        other._as_type = None if data_type is None else to_pa_datatype(data_type)
        return other

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        if self.value is None:
            return self.as_type
        return self.as_type or to_pa_datatype(type(self.value))

    def _uuid_keys(self) -> List[Any]:
        return [self.value]


class _FuncExpr(ColumnExpr):
    def __init__(
        self,
        func: str,
        *args: Any,
        arg_distinct: bool = False,
        **kwargs: Any,
    ):
        self._distinct = arg_distinct
        self._func = func
        self._args = list(args)
        self._kwargs = dict(kwargs)
        super().__init__()

    @property
    def body_str(self) -> str:
        def to_str(v: Any):
            if isinstance(v, str):
                return f"'{v}'"
            if isinstance(v, bool):
                return "TRUE" if v else "FALSE"
            return str(v)

        a1 = [to_str(x) for x in self.args]
        a2 = [k + "=" + to_str(v) for k, v in self.kwargs.items()]
        args = ",".join(a1 + a2)
        distinct = "DISTINCT " if self.is_distinct else ""
        return f"{self.func}({distinct}{args})"

    @property
    def func(self) -> str:
        return self._func

    @property
    def is_distinct(self) -> bool:
        return self._distinct

    @property
    def args(self) -> List[Any]:
        return self._args

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self._kwargs

    def alias(self, as_name: str) -> ColumnExpr:
        other = self._copy()
        other._as_name = as_name
        other._distinct = self.is_distinct
        other._as_type = self.as_type
        return other

    def cast(self, data_type: Any) -> ColumnExpr:
        other = self._copy()
        other._as_name = self.as_name
        other._distinct = self.is_distinct
        other._as_type = None if data_type is None else to_pa_datatype(data_type)
        return other

    def _copy(self) -> "_FuncExpr":
        return _FuncExpr(self.func, *self.args, **self.kwargs)

    def _uuid_keys(self) -> List[Any]:
        return [self.func, self.is_distinct, self.args, self.kwargs]


class _UnaryOpExpr(_FuncExpr):
    def __init__(self, op: str, column: ColumnExpr, arg_distinct: bool = False):
        super().__init__(op, column, arg_distinct=arg_distinct)

    @property
    def col(self) -> ColumnExpr:
        return self.args[0]

    @property
    def op(self) -> str:
        return self.func

    def infer_alias(self) -> ColumnExpr:
        return (
            self
            if self.output_name != ""
            else self.alias(self.col.infer_alias().output_name)
        )

    def _copy(self) -> _FuncExpr:
        return _UnaryOpExpr(self.op, self.col)


class _InvertOpExpr(_UnaryOpExpr):
    def _copy(self) -> _FuncExpr:
        return _InvertOpExpr(self.op, self.col)

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        if self.as_type is not None:
            return self.as_type
        tp = self.col.infer_type(schema)
        if pa.types.is_signed_integer(tp) or pa.types.is_floating(tp):
            return tp
        return None


class _NotOpExpr(_UnaryOpExpr):
    def _copy(self) -> _FuncExpr:
        return _NotOpExpr(self.op, self.col)

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        if self.as_type is not None:
            return self.as_type
        tp = self.col.infer_type(schema)
        if pa.types.is_boolean(tp):
            return tp
        return None


class _BinaryOpExpr(_FuncExpr):
    def __init__(self, op: str, left: Any, right: Any, arg_distinct: bool = False):
        super().__init__(op, _to_col(left), _to_col(right), arg_distinct=arg_distinct)

    @property
    def left(self) -> ColumnExpr:
        return self.args[0]

    @property
    def right(self) -> ColumnExpr:
        return self.args[1]

    @property
    def op(self) -> str:
        return self.func

    def _copy(self) -> _FuncExpr:
        return _BinaryOpExpr(self.op, self.left, self.right)


class _BoolBinaryOpExpr(_BinaryOpExpr):
    def _copy(self) -> _FuncExpr:
        return _BoolBinaryOpExpr(self.op, self.left, self.right)

    def infer_type(self, schema: Schema) -> Optional[pa.DataType]:
        return self.as_type or pa.bool_()



import json
from typing import Any, Tuple

import pandas as pd

from fugue import Outputter
from fugue.extensions import namespace_candidate, parse_outputter

from ._ext import Visualize


@parse_outputter.candidate(namespace_candidate("viz", lambda x: isinstance(x, str)))
def _parse_pandas_plot(obj: Tuple[str, str]) -> Outputter:
    return _PandasVisualize(obj[1])


class _PandasVisualize(Visualize):
    def __init__(self, func: str) -> None:
        super().__init__(func)
        if func != "plot":
            getattr(pd.DataFrame.plot, func)  # ensure the func exists

    def _plot(self, df: pd.DataFrame) -> None:
        params = dict(self.params)
        if len(self.partition_spec.partition_by) > 0:
            keys = df[self.partition_spec.partition_by].head(1).to_dict("records")[0]
            kt = json.dumps(keys)[1:-1]
            if "title" in params:
                params["title"] = params["title"] + " -- " + kt
            else:
                params["title"] = kt
            df = df.drop(self.partition_spec.partition_by, axis=1)
        func = self._get_func(df)
        func(**params)

    def _get_func(self, df: pd.DataFrame) -> Any:
        if self._func == "plot":
            return df.plot
        return getattr(df.plot, self._func)
